#! python3  # noqa E265

"""
    Usage from the repo root folder:

    .. code-block:: bash
        # for whole tests
        python -m unittest tests.test_utils_ini_custom_parser
        # for specific test
        python -m unittest tests.test_utils_ini_custom_parser.TestCustomConfigParser.test_file_path_storing
"""

# standard
import unittest
from pathlib import Path

# project
from qgis_deployment_toolbelt.utils.ini_parser_with_path import CustomConfigParser

# ############################################################################
# ########## Classes #############
# ################################


class TestCustomConfigParser(unittest.TestCase):
    """Test extended ConfigParser that store the initial ini file pathmodule."""

    def test_file_path_storing(self):
        # temporary file
        test_file_path = Path("tests/fixtures/tmp/test_custom_config_parser.ini")
        test_file_path.parent.mkdir(parents=True, exist_ok=True)
        test_file_path.write_text("[Section1]\noption1 = value1\n")

        # read and get back the initial file path
        parser = CustomConfigParser()
        parser.read(test_file_path)

        self.assertEqual(parser.get_initial_file_path(), test_file_path)

        # clean up
        test_file_path.unlink()


# ############################################################################
# ####### Stand-alone run ########
# ################################
if __name__ == "__main__":
    unittest.main()



#! python3  # noqa E265

"""
    Usage from the repo root folder:

    .. code-block:: bash
        # for whole tests
        python -m unittest tests.test_utils_images_size
        # for specific test
        python -m unittest tests.test_utils_images_size.TestUtilsImagesSizeChecker.get_svg_size
"""


# standard library
import unittest
from pathlib import Path
from tempfile import TemporaryDirectory

# 3rd party
from PIL import Image

# project
from qgis_deployment_toolbelt.__about__ import __title_clean__, __version__
from qgis_deployment_toolbelt.utils.check_image_size import (
    check_image_dimensions,
    get_image_size,
)

# ############################################################################
# ########## Classes #############
# ################################


class TestUtilsImagesSizeChecker(unittest.TestCase):
    """Test package utilities."""

    # -- Standard methods --------------------------------------------------------
    @classmethod
    def setUpClass(cls) -> None:
        """Set up test."""

        cls.img_tmp_folder = TemporaryDirectory(
            prefix=f"{__title_clean__}_{__version__}_"
        )

        # create a temporary image
        cls.img_black_800x600 = Path(cls.img_tmp_folder.name, "img_black_800x600.jpg")
        width = 800
        height = 600
        image = Image.new("RGB", (width, height), "black")
        image.save(cls.img_black_800x600)

    @classmethod
    def tearDownClass(cls):
        """Executed after each test."""
        cls.img_tmp_folder.cleanup()

    # -- TESTS ---------------------------------------------------------
    def test_svg_size_with_dimensions(self):
        """Test svg size retriever."""
        svg_with_dimensions_attributes = Path(
            "tests/fixtures/miscellaneous/sample_with_dimensions_attributes.svg"
        )

        self.assertTrue(svg_with_dimensions_attributes.is_file())
        # svg with dimensions set
        svg_size = get_image_size(image_filepath=svg_with_dimensions_attributes)

        self.assertIsInstance(svg_size, tuple)
        self.assertIsInstance(svg_size[0], int)
        self.assertIsInstance(svg_size[1], int)
        self.assertEqual(svg_size[0], 853)
        self.assertEqual(svg_size[1], 568)

    def test_svg_size_without_dimensions(self):
        """Test svg size retriever."""
        svg_without_dimensions_attributes = Path(
            "tests/fixtures/miscellaneous/sample_without_dimensions_attributes.svg"
        )
        self.assertTrue(svg_without_dimensions_attributes.is_file())
        svg_size = get_image_size(image_filepath=svg_without_dimensions_attributes)

        self.assertIsNone(svg_size)

    def test_get_image_dimensions(self):
        img_800x600 = get_image_size(self.img_black_800x600)
        self.assertIsInstance(img_800x600, tuple)
        self.assertEqual(img_800x600[0], 800)
        self.assertEqual(img_800x600[1], 600)

    def test_check_image_dimensions(self):
        """Test image dimensions checker."""
        self.assertTrue(
            check_image_dimensions(
                image_filepath=self.img_black_800x600, max_width=801, max_height=601
            )
        )

        self.assertFalse(
            check_image_dimensions(
                image_filepath=self.img_black_800x600, max_width=300, max_height=601
            )
        )

        self.assertFalse(
            check_image_dimensions(
                image_filepath=self.img_black_800x600, max_width=2000, max_height=200
            )
        )

        self.assertFalse(
            check_image_dimensions(
                image_filepath=self.img_black_800x600, max_width=300, max_height=500
            )
        )

        self.assertFalse(
            check_image_dimensions(
                image_filepath=self.img_black_800x600,
                max_width=300,
                max_height=500,
                allowed_images_extensions=(".png", ".webp"),
            )
        )


# ############################################################################
# ####### Stand-alone run ########
# ################################
if __name__ == "__main__":
    unittest.main()



#! python3  # noqa: E265

"""
    Check image size using pure Python <https://github.com/shibukawa/imagesize_py>.

    Author: Julien Moura (https://github.com/guts)
"""

# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import logging
import xml.etree.ElementTree as ET
from decimal import Decimal
from pathlib import Path

# 3rd party
import imagesize

# #############################################################################
# ########## Globals ###############
# ##################################

# logs
logger = logging.getLogger(__name__)


# #############################################################################
# ########## Functions #############
# ##################################


def get_image_size(image_filepath: Path) -> tuple[int, int]:
    """Get image dimensions as a tuple (width,height). Return None in case of error.

    :param Path image_filepath: path to the image

    :return Tuple[int, int]: dimensions tuple (width,height)
    """
    # handle SVG
    if image_filepath.suffix.lower() == ".svg":
        svg_size = get_svg_size(image_filepath)
        if not svg_size:
            return None
        else:
            return svg_size

    # get image dimensions
    try:
        return imagesize.get(image_filepath)
    except ValueError as exc:
        logging.error(f"Invalid image: {image_filepath.resolve()}. Trace: {exc}")
    except Exception as exc:
        logging.error(
            f"Something went wrong reading the image: {image_filepath.resolve()}. Trace: {exc}"
        )

    return None


def get_svg_size(image_filepath: Path) -> tuple[int, int]:
    """Extract SVG width and height from a SVG file and convert them into integers. \
    Relevant and working only if the file root has width and height attributes.

    :param Path image_filepath: path to the svg file

    :return Tuple[int, int]: tuple of dimensions as integers (width,height)
    """
    try:
        tree = ET.parse(image_filepath)
        root = tree.getroot()
    except Exception as err:
        logger.error(f"Unable to open SVG file as XML: {image_filepath}. Trace: {err}")
        return None

    try:
        return int(Decimal(root.attrib["width"])), int(Decimal(root.attrib["height"]))
    except Exception as err:
        logger.warning(
            "Unable to determine image dimensions from width/height "
            f"attributes: {image_filepath}. It migh be infinitely scalable. Trace: {err}"
        )
        return None


def check_image_dimensions(
    image_filepath: Path,
    min_width: int = 500,
    max_width: int = 600,
    min_height: int = 250,
    max_height: int = 350,
    allowed_images_extensions: tuple = (".jpg", ".jpeg", ".png", ".svg"),
) -> bool:
    """Check input image dimensions against passed limits.

    :param Path image_filepath: path to the image to check
    :param int min_width: minimum width, defaults to 500
    :param int max_width: maximum width, defaults to 600
    :param int min_height: minimum height, defaults to 250
    :param int max_height: maximum height, defaults to 350

    :return bool: True if image dimensions are inferior
    """

    if image_filepath.suffix.lower() not in allowed_images_extensions:
        logger.error(
            f"Image extension {image_filepath.suffix.lower()} is not one of "
            f"supported: {allowed_images_extensions}"
        )
        return None

    image_dimensions = get_image_size(image_filepath=image_filepath)
    if not image_dimensions:
        logger.info(
            f"Unable to determine image dimensions ({image_filepath.resolve()}), "
            "so unable to check it it complies with limits."
        )
        return None

    return all(
        dim <= limit for dim, limit in zip(image_dimensions, (max_width, max_height))
    )


# #############################################################################
# ##### Stand alone program ########
# ##################################

if __name__ == "__main__":
    """Standalone execution."""
    svg_path = Path(
        "tests/fixtures/miscellaneous/sample_with_dimensions_attributes.svg"
    )
    assert svg_path.is_file()
    print(get_svg_size(image_filepath=svg_path))
    print(get_image_size(image_filepath=svg_path))

    svg_path = Path(
        "tests/fixtures/miscellaneous/sample_without_dimensions_attributes.svg"
    )
    assert svg_path.is_file()
    print(get_svg_size(image_filepath=svg_path))



import configparser
from os import PathLike
from pathlib import Path
from typing import Any


class CustomConfigParser(configparser.ConfigParser):
    """CustomConfigParser extends configparser.ConfigParser to include
    functionality for storing the paths of the INI files being read.

    Attributes:
        initial_file_path (str | PathLike | None): path of the initial ini file from
            which the configurations were loaded.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize the CustomConfigParser with the same arguments as
        configparser.ConfigParser.
        """
        super().__init__(*args, **kwargs)
        self.initial_file_path: str | PathLike | None = None

    def read(self, ini_filepath: str | PathLike, *args: Any, **kwargs: Any) -> None:
        """Read and parse filenames, storing the file paths.

        Args:
            ini_filepath ( str | PathLike): The paths of
                the INI files to be read.
        """
        super().read(ini_filepath, *args, **kwargs)
        self.initial_file_path = ini_filepath

    def get_initial_file_path(self) -> Path | None:
        """Get the stored file path of the initial INI file as pathlib.Path object.

        Returns:
            Path: path to the initial INI file that have been read.
        """
        if self.initial_file_path is not None:
            return Path(self.initial_file_path)
        return None



#! python3  # noqa E265

"""
    Usage from the repo root folder:

    .. code-block:: bash
        # for whole tests
        python -m unittest tests.test_qdt_profile_object
        # for specific test
        python -m unittest tests.test_qdt_profile_object.TestQdtProfile.test_profile_load_from_json_basic
"""

# standard
import unittest
from pathlib import Path

# project
from qgis_deployment_toolbelt.profiles.qdt_profile import QdtProfile

# ############################################################################
# ########## Classes #############
# ################################


class TestQdtProfile(unittest.TestCase):
    """Test QDT profile abstraction class."""

    # -- Standard methods --------------------------------------------------------
    @classmethod
    def setUpClass(cls):
        """Executed when module is loaded before any test."""
        cls.good_profiles_files = sorted(
            Path("tests/fixtures/").glob("profiles/good_*.json")
        )

    def test_profile_load_from_json_basic(self):
        """Test profile loading from JSON."""
        for i in self.good_profiles_files:
            qdt_profile = QdtProfile.from_json(profile_json_path=i)
            self.assertIsInstance(qdt_profile, QdtProfile)

            # attributes types
            self.assertIsInstance(qdt_profile.name, str)
            self.assertIsInstance(qdt_profile.version, str)
            self.assertIsInstance(qdt_profile.plugins, list)

    def test_profile_load_from_json_with_parent_folder(self):
        """Test profile loading from JSON specifying parent folder."""
        for i in self.good_profiles_files:
            qdt_profile = QdtProfile.from_json(i, i.parent)
            self.assertIsInstance(qdt_profile, QdtProfile)

            # attributes types
            self.assertIsInstance(qdt_profile.name, str)
            self.assertIsInstance(qdt_profile.version, str)
            self.assertIsInstance(qdt_profile.plugins, list)

    def test_profile_load_from_json_complete(self):
        """Test profile loading from JSON."""
        for i in filter(lambda x: "complete" in x.name, self.good_profiles_files):
            qdt_profile = QdtProfile.from_json(i, i.parent)
            self.assertIsInstance(qdt_profile, QdtProfile)

            # attributes types
            self.assertIsInstance(qdt_profile.name, str)
            self.assertIsInstance(qdt_profile.alias, str)
            self.assertIsInstance(qdt_profile.folder, Path)
            self.assertIsInstance(qdt_profile.splash, (str, Path))
            self.assertIsInstance(qdt_profile.version, str)

            # attributes values
            self.assertEqual(i.parent.resolve(), qdt_profile.folder)

    def test_profile_versions_comparison_semver(self):
        """Test profile compare versions semver"""
        profile_v1: QdtProfile = QdtProfile(
            alias="Unit Test lesser",
            name="unit_test_1",
            version="1.0.0",
        )

        profile_v2: QdtProfile = QdtProfile(
            alias="Unit Test lesser",
            name="unit_test_1",
            version="1.1.0",
        )

        profile_v3: QdtProfile = QdtProfile(
            alias="Unit Test lesser",
            name="unit_test_1",
            version="3.1.0",
        )

        self.assertTrue(profile_v1.is_older_than(profile_v2.version))
        self.assertTrue(profile_v1.is_older_than(profile_v2))
        self.assertTrue(profile_v1.is_older_than(profile_v3.version))
        self.assertTrue(profile_v1.is_older_than(profile_v3))
        self.assertTrue(profile_v2.is_older_than(profile_v3.version))
        self.assertTrue(profile_v2.is_older_than(profile_v3))
        self.assertFalse(profile_v2.is_older_than(profile_v1))


# ############################################################################
# ####### Stand-alone run ########
# ################################
if __name__ == "__main__":
    unittest.main()



#! python3  # noqa: E265

"""
    Test CLI's clean command.

    Author: Julien Moura (Oslandia)
"""

# #############################################################################
# ########## Libraries #############
# ##################################


# standard library
import unittest

# 3rd party library
import pytest
from validators import url

# module to test
from qgis_deployment_toolbelt import cli
from qgis_deployment_toolbelt.__about__ import __uri_repository__
from qgis_deployment_toolbelt.commands.upgrade import (
    get_download_url_for_os,
    get_latest_release,
    replace_domain,
)
from qgis_deployment_toolbelt.constants import SUPPORTED_OPERATING_SYSTEMS_CODENAMES

# #############################################################################
# ######## Functions ###############
# ##################################


def test_cli_upgrade_check_only(capsys):
    """Test CLI upgrade check only."""
    with pytest.raises(SystemExit):
        cli.main(["upgrade", "-c"])

    _, err = capsys.readouterr()

    assert err == ""


def test_cli_upgrade_download(capsys):
    """Test CLI upgrade ."""
    with pytest.raises(SystemExit):
        cli.main(["upgrade", "-n", "-w", "tests/"])

    _, err = capsys.readouterr()

    assert err == ""


# ############################################################################
# ########## Classes #############
# ################################


class TestUpgradeUtils(unittest.TestCase):
    """Test upgrade utilities."""

    def test_release_assets_download_links(self):
        """Test utils to retrieve latest version download links."""
        api_url_from_repo_url = replace_domain(
            url=__uri_repository__, new_domain="api.github.com/repos"
        )
        self.assertTrue(url(api_url_from_repo_url))
        self.assertIsInstance(api_url_from_repo_url, str)

        latest_release = get_latest_release(api_url_from_repo_url)
        self.assertIsInstance(latest_release, dict)
        self.assertTrue("assets" in latest_release)

        dl_hyperlinks = [
            get_download_url_for_os(latest_release.get("assets"), override_opersys=os)[
                0
            ]
            for os in SUPPORTED_OPERATING_SYSTEMS_CODENAMES
        ]
        self.assertTrue(len(dl_hyperlinks), 3)

        self.assertTrue(
            all(
                [
                    dl_url.startswith("https")
                    for dl_url in dl_hyperlinks
                    if isinstance(dl_url, str)
                ]
            )
        )


# #############################################################################
# ######## Standalone ##############
# ##################################
if __name__ == "__main__":
    unittest.main()



#! python3  # noqa E265

"""Usage from the repo root folder:

    .. code-block:: python

        # for whole test
        python -m unittest tests.test_tree_qdt-files_json
        # for specific
        python -m unittest tests.test_tree_qdt-files_json.TestTreeQdtFilesReader.test_load_tree_json_files
"""

# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import json
import unittest
from pathlib import Path

# module target
from qgis_deployment_toolbelt.utils.tree_files_reader import tree_to_download_list

# #############################################################################
# ########## Classes ###############
# ##################################


class TestTreeQdtFilesReader(unittest.TestCase):
    """Test module."""

    # -- Standard methods --
    @classmethod
    def setUpClass(cls):
        """Executed when module is loaded before any test."""
        cls.tree_qdt_files = sorted(
            Path("tests/fixtures/").glob("treefiles/qdt-files*.json")
        )

    # -- Tests methods --
    def test_load_tree_json_files(self):
        """Test tree files loading and parsing."""
        self.assertGreaterEqual(len(self.tree_qdt_files), 1)
        for tree_file in self.tree_qdt_files:
            print(f"Processing {tree_file}")

            with tree_file.open(mode="r", encoding="utf-8") as in_json:
                tree_data = json.load(in_json)

            li_files_to_download = tree_to_download_list(tree_array=tree_data)
            # check type
            self.assertIsInstance(li_files_to_download, list)
            self.assertTrue(all([isinstance(f, str) for f in li_files_to_download]))



#! python3  # noqa: E265

"""
    Sub-command in charge of checking if new versions are available.

    Author: Julien M. (https://github.com/guts)
"""

# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import argparse
import logging
import sys
from collections.abc import Iterable
from os import getenv
from pathlib import Path
from sys import platform as opersys
from urllib.parse import urlsplit, urlunsplit

# 3rd party library
import requests
from packaging.version import Version

# submodules
from qgis_deployment_toolbelt.__about__ import (
    __package_name__,
    __title__,
    __title_clean__,
    __uri_repository__,
)
from qgis_deployment_toolbelt.__about__ import __version__ as actual_version
from qgis_deployment_toolbelt.utils.bouncer import (
    exit_cli_error,
    exit_cli_normal,
    exit_cli_success,
)
from qgis_deployment_toolbelt.utils.file_downloader import download_remote_file_to_local
from qgis_deployment_toolbelt.utils.proxies import get_proxy_settings
from qgis_deployment_toolbelt.utils.str2bool import str2bool

# #############################################################################
# ########## Globals ###############
# ##################################

logger = logging.getLogger(__name__)

# #############################################################################
# ####### Functions ###############
# #################################


def get_download_url_for_os(
    release_assets: list, override_opersys: str | None = None
) -> tuple[str | None, str | None]:
    """Parse list of a GitHub release assets and return the appropriate download URL \
        for the current operating system.

    Args:
        release_assets (list): list of assets
        override_opersys (str, optional): override current operating system code. Useful
            to get a download URL for a specific OS. Defaults to None.

    Returns:
        tuple[str, str]: tuple containing asset download URL (browser_download_url) and
            content-type (barely defined)
    """
    opersys_code = opersys
    if override_opersys is not None:
        opersys_code = override_opersys

    for asset in release_assets:
        if opersys_code == "win32" and "Windows" in asset.get("name"):
            return asset.get("browser_download_url"), asset.get("content-type")
        elif opersys_code == "linux" and "Ubuntu" in asset.get("name"):
            return asset.get("browser_download_url"), asset.get("content-type")
        elif opersys_code == "darwin" and "MacOS" in asset.get("name"):
            return asset.get("browser_download_url"), asset.get("content-type")
        else:
            continue

    return None, None


def get_latest_release(api_repo_url: str) -> dict | None:
    """Get latest release from GitHub public API.

    Args:
        api_repo_url (str): API URL with the owner and repository set

    Returns:
        dict: GitHub release object
    """

    request_url = f"{api_repo_url}releases/latest"

    # headers
    headers = {
        "content-type": "application/vnd.github+json",
        "User-Agent": f"{__title_clean__}/{actual_version}",
    }
    if getenv("GITHUB_TOKEN"):
        logger.debug(
            f"Using authenticated request to GH API: {getenv('GITHUB_TOKEN')[:9]}****"
        )
        headers["Authorization"] = f"Bearer {getenv('GITHUB_TOKEN')}"

    try:
        release_info = None
        req = requests.get(
            url=request_url,
            headers=headers,
            proxies=get_proxy_settings(url=request_url),
        )
        req.raise_for_status()
        release_info = req.json()
        return release_info
    except Exception as err:
        logger.error(err)
        if isinstance(err, Iterable) and "rate limit exceeded" in str(err):
            logger.error(
                "Rate limit of GitHub API exeeded. Try again later (generally "
                "in 15 minutes) or set GITHUB_TOKEN as environment variable with a "
                "personal token."
            )
        return None


def replace_domain(url: str, new_domain: str = "api.github.com/repos") -> str:
    """Replaces the domain of an URL with a new domain.

    Args:
        url (str): The original URL.
        new_domain (str, optional): The new domain to replace the original domain with. Defaults
            to "api.github.com/repos".

    Returns:
        str: The URL with the new domain.
    """
    split_url = urlsplit(url)
    split_url = split_url._replace(netloc=new_domain)
    new_url = urlunsplit(split_url)
    return new_url


# ############################################################################
# ########## CLI #################
# ################################


def parser_upgrade(
    subparser: argparse.ArgumentParser,
) -> argparse.ArgumentParser:
    """Set the argument parser subcommand.

    Args:
        subparser (argparse.ArgumentParser): parser to set up

    Returns:
        argparse.ArgumentParser: parser ready to use
    """
    subparser.add_argument(
        "-c",
        "--check-only",
        help="Only check if a new version is available. No download.",
        default=str2bool(getenv("QDT_UPGRADE_CHECK_ONLY", False)),
        action="store_true",
        dest="opt_only_check",
    )

    subparser.add_argument(
        "-n",
        "--dont-show-release-notes",
        help="Display release notes.",
        default=str2bool(getenv("QDT_UPGRADE_DISPLAY_RELEASE_NOTES", True)),
        action="store_false",
        dest="opt_show_release_notes",
    )

    subparser.add_argument(
        "-w",
        "--where",
        help="Folder to store the downloaded file.",
        default=getenv("QDT_UPGRADE_DOWNLOAD_FOLDER", "./"),
        type=Path,
        dest="local_download_folder",
    )

    subparser.set_defaults(func=run)

    return subparser


# ############################################################################
# ########## MAIN ################
# ################################


def run(args: argparse.Namespace):
    """Run the sub command logic.

    Check if a new version of the CLI is available and download it if needed.

    Args:
        args (argparse.Namespace): arguments passed to the subcommand
    """
    logger.debug(f"Running {args.command} with {args}")

    # build API URL from repository
    api_url = replace_domain(url=__uri_repository__)

    # get latest release as dictionary
    latest_release: dict | None = get_latest_release(api_repo_url=api_url)

    if not isinstance(latest_release, dict):
        exit_cli_error(
            f"Unable to retrieve latest release {type(latest_release)} from {api_url}."
        )

    # compare it
    latest_version: str = latest_release.get("tag_name")
    if Version(version=actual_version) < Version(version=latest_version):
        print(f"A newer version is available: {latest_version}")
        if args.opt_show_release_notes:
            print(latest_release.get("body"))
        if args.opt_only_check:
            exit_cli_normal(
                f"A newer version is available: {latest_version}. No download because "
                "of option check-only enabled.",
                abort=True,
            )
    else:
        exit_msg = f"You already have the latest released version: {latest_version}."
        print(exit_msg)
        exit_cli_normal(
            message=exit_msg,
            abort=True,
        )

    # -- DOWNLOAD ------------------------------------------------------------
    # check if we are in frozen mode (typically PyInstaller) or as "normal" Python
    if not (getattr(sys, "frozen", False) and hasattr(sys, "_MEIPASS")):
        logger.debug("Running in a normal Python process.")
        print(
            "\n\nTo get the latest version, run (adapt command to your environment):"
            f"\n\npython -m pip install -U {__package_name__}"
        )
        sys.exit(0)

    print(f"Downloading newer version of executable for {opersys}: {latest_version}")

    # select remote download URL
    if release_asset_for_os := get_download_url_for_os(latest_release.get("assets")):
        remote_url, remote_content_type = release_asset_for_os
    else:
        exit_cli_error(f"Unable to identify an appropriate download URL for {opersys}.")

    # handle empty content-type
    if remote_content_type is None:
        remote_content_type = "application/octet-stream"

    # destination local file
    dest_filepath = Path(
        args.local_download_folder, Path(urlsplit(remote_url).path).name
    )

    # download it
    logger.info(
        f"Downloading new version '{latest_version}' from {remote_url} to {dest_filepath}"
    )
    try:
        download_remote_file_to_local(
            remote_url_to_download=remote_url,
            local_file_path=dest_filepath,
            content_type=remote_content_type,
            use_stream=str2bool(getenv("QDT_STREAMED_DOWNLOADS", True)),
        )
    except Exception as err:
        exit_cli_error(f"Download new version failed. Trace: {err}")

    exit_cli_success(f"New version of {__title__} downloaded here: {dest_filepath}.")


# #############################################################################
# ##### Stand alone program ########
# ##################################
if __name__ == "__main__":
    """Standalone execution."""
    latest_release = get_latest_release(
        replace_domain(url=__uri_repository__, new_domain="api.github.com/repos")
    )
    print(
        latest_release.keys(),
        latest_release.get("assets_url"),
        # latest_release.get("assets"),
    )

    dl_link_linux, dl_link_macos, dl_link_windows = (
        get_download_url_for_os(latest_release.get("assets"), override_opersys=os)[0]
        for os in ["linux", "darwin", "win32"]
    )
    print(dl_link_linux, dl_link_macos, dl_link_windows)



#! python3  # noqa E265

"""
    Usage from the repo root folder:

    .. code-block:: bash
        # for whole tests
        python -m unittest tests.test_utils_formatters
        # for specific test
        python -m unittest tests.test_utils_formatters.TestUtilsFormatters.test_convert_octets
"""

# standard library
import unittest

# project
from qgis_deployment_toolbelt.utils.formatters import (
    convert_octets,
    url_ensure_trailing_slash,
)

# ############################################################################
# ########## Classes #############
# ################################


class TestUtilsFormatters(unittest.TestCase):
    """Test package utilities."""

    def test_convert_octets(self):
        """Test file size formatter."""
        self.assertEqual(
            convert_octets(1024),
            "1.0 Ko",
        )

        self.assertEqual(
            convert_octets(2097152),
            "2.0 Mo",
        )

    def test_url_ensure_trailing_slash(self):
        """Test URL trailing slash."""
        self.assertEqual(
            url_ensure_trailing_slash(
                in_url="https://qgis-deployment.github.io/qgis-deployment-toolbelt-cli"
            ),
            "https://qgis-deployment.github.io/qgis-deployment-toolbelt-cli/",
        )

        self.assertEqual(
            url_ensure_trailing_slash(
                in_url="https://qgis-deployment.github.io/qgis-deployment-toolbelt-cli/"
            ),
            "https://qgis-deployment.github.io/qgis-deployment-toolbelt-cli/",
        )


# ############################################################################
# ####### Stand-alone run ########
# ################################
if __name__ == "__main__":
    unittest.main()



#! python3  # noqa: E265

"""
    Plugin object model.

    Author: Julien Moura (https://github.com/guts)
"""


# #############################################################################
# ########## Libraries #############
# ##################################

# special
from __future__ import annotations

# Standard library
import configparser
import logging
import zipfile
from dataclasses import dataclass, fields
from enum import Enum
from os.path import expanduser, expandvars
from pathlib import Path
from urllib.parse import quote, urlsplit, urlunsplit

# 3rd party
from packaging.version import InvalidVersion, Version

# package
from qgis_deployment_toolbelt.utils.check_path import check_path
from qgis_deployment_toolbelt.utils.slugger import sluggy

# #############################################################################
# ########## Globals ###############
# ##################################

# logs
logger = logging.getLogger(__name__)

# #############################################################################
# ########## Classes ###############
# ##################################


class QgisPluginLocation(Enum):
    local = 1
    remote = 2


@dataclass
class QgisPlugin:
    """Model describing a QGIS plugin."""

    # optional mapping on attributes names.
    # Structure: {attribute_name_in_output_object: attribute_name_from_input_file} # noqa: E800
    ATTR_MAP = {
        "location": "type",
        "qgis_maximum_version": "qgisMaximumVersion",
        "qgis_minimum_version": "qgisMinimumVersion",
    }

    OFFICIAL_REPOSITORY_URL_BASE = "https://plugins.qgis.org/"
    OFFICIAL_REPOSITORY_XML = "https://plugins.qgis.org/plugins/plugins.xml"

    name: str
    folder_name: str = None
    location: QgisPluginLocation = "remote"
    official_repository: bool = None
    plugin_id: int = None
    qgis_maximum_version: str = None
    qgis_minimum_version: str = None
    repository_url_xml: str = None
    url: str = None
    version: str = "latest"

    @classmethod
    def from_dict(cls, input_dict: dict) -> QgisPlugin:
        """Create object from a dictionary.

        Args:
            input_dict (dict): input dictionary

        Returns:
            QgisPlugin: instanciated object
        """
        # map attributes names
        for k, v in cls.ATTR_MAP.items():
            if v.lower() in input_dict.keys():
                input_dict[k] = input_dict.pop(v.lower(), None)

        # official repository autodetection
        if input_dict.get("repository_url_xml") == cls.OFFICIAL_REPOSITORY_XML:
            input_dict["official_repository"] = True
        elif (
            input_dict.get("url")
            and isinstance(input_dict.get("url"), str)
            and input_dict.get("url").startswith(cls.OFFICIAL_REPOSITORY_URL_BASE)
        ):
            input_dict["official_repository"] = True
            input_dict["repository_url_xml"] = cls.OFFICIAL_REPOSITORY_XML
        else:
            pass

        # URL auto build
        if input_dict.get("official_repository") is True and not input_dict.get("url"):
            input_dict["url"] = (
                f"{cls.OFFICIAL_REPOSITORY_URL_BASE}"
                f"plugins/{input_dict.get('folder_name') or input_dict.get('name')}/"
                f"version/{input_dict.get('version')}/download/"
            )
            input_dict["repository_url_xml"] = cls.OFFICIAL_REPOSITORY_XML
            input_dict["location"] = "remote"

        # remove keys which are not in object attributes
        attributes_names = [f.name for f in fields(cls)]
        for k in list(input_dict):
            if k not in attributes_names:
                del input_dict[k]

        # return new instance with loaded object
        return cls(
            **input_dict,
        )

    @classmethod
    def from_plugin_folder(cls, input_plugin_folder: Path) -> QgisPlugin:
        """Create object from a QGIS plugin folder. Must contain a metadata.txt file.

        Args:
            input_plugin_folder (Path): path to the folder containgin a QGIS plugin

        Returns:
            QgisPlugin: instanciated object
        """
        # check that input path is a folder
        check_path(
            input_path=input_plugin_folder,
            must_exists=True,
            must_be_a_folder=True,
            must_be_a_file=False,
            must_be_readable=True,
        )
        # check if the folder contains a metadata.txt file
        plugin_metadata_txt = input_plugin_folder / "metadata.txt"
        check_path(
            input_path=plugin_metadata_txt,
            must_be_a_file=True,
            must_be_readable=True,
            must_exists=True,
        )

        # read it
        with plugin_metadata_txt.open(encoding="UTF-8") as config_file:
            config = configparser.ConfigParser(strict=False)
            config.read_file(config_file)
        plugin_md_as_dict = {k: v for k, v in config.items(section="general")}

        # add folder name
        plugin_md_as_dict["folder_name"] = input_plugin_folder.name

        return cls.from_dict(plugin_md_as_dict)

    @classmethod
    def from_zip(cls, input_zip_path: Path) -> QgisPlugin:
        """Create object from a ZIP file.

        Args:
            input_zip_path (Path): filepath of the input zip

        Returns:
            QgisPlugin: instanciated object
        """
        with zipfile.ZipFile(file=input_zip_path) as zf:
            # find the metadata.txt file
            for i in zf.infolist():
                if not i.is_dir() and i.filename.split("/")[1] == "metadata.txt":
                    break

            # open and read it
            zip_path = zipfile.Path(zf)
            metadata_file = zip_path / i.filename
            plugin_folder_name = metadata_file.parent.name
            with metadata_file.open(encoding="UTF-8") as config_file:
                config = configparser.ConfigParser(strict=False)
                config.read_file(config_file)

        plugin_md_as_dict = {k: v for k, v in config.items(section="general")}

        # add folder name
        plugin_md_as_dict["folder_name"] = plugin_folder_name

        return cls.from_dict(plugin_md_as_dict)

    @property
    def download_url(self) -> str:
        """Try to guess download URL if it's not set during the object init.

        Returns:
            str: download URL
        """
        if self.url:
            return quote(self.url, safe="/:")
        elif self.repository_url_xml and self.folder_name and self.version:
            split_url = urlsplit(self.repository_url_xml)
            new_url = split_url._replace(path=split_url.path.replace("plugins.xml", ""))
            return f"{urlunsplit(new_url)}{self.folder_name}.{self.version}.zip"
        elif self.repository_url_xml and self.name and self.version:
            split_url = urlsplit(self.repository_url_xml)
            new_url = split_url._replace(path=split_url.path.replace("plugins.xml", ""))
            return f"{urlunsplit(new_url)}{self.name}.{self.version}.zip"
        else:
            return None

    @property
    def id_with_version(self) -> str:
        """Unique identifier using plugin_id (if set) and name + version slugified.

        Returns:
            str: plugin identifier meant to be unique per version
        """
        if self.plugin_id:
            return f"{self.plugin_id}_{sluggy(self.name)}_{sluggy(self.version.replace('.', '-'))}"
        else:
            return f"{sluggy(self.name)}_{sluggy(self.version.replace('.', '-'))}"

    @property
    def installation_folder_name(self) -> str:
        """Name of the folder when the plugin is installed into QGIS/profile/python/plugins/. \
            If not clearly specified intot the profile.json, it tries to extract it from \
            the download URL. As final fallback, it returns the slufigied plugin name.

        Returns:
            str: plugin folder name
        """
        if self.folder_name:
            return self.folder_name
        elif self.download_url:
            try:
                folder_name_from_url = urlsplit(self.download_url).path.split("/")[2]
                self.folder_name = folder_name_from_url
                return folder_name_from_url
            except Exception as err:
                logger.error(
                    f"Plugin {self.name} - Determine plugin folder name from download"
                    f"URL failed. Please specify it into profile.json file. Trace: {err}"
                )
                return sluggy(self.name)
        else:
            return sluggy(self.name)

    def is_older_than(self, version_to_compare: str | QgisPlugin) -> bool:
        """Determine if the actual object version is older than the given version to \
            compare.

        Args:
            version_to_compare (Union[str, QgisPlugin]): given version to compare with object version

        Returns:
            bool: True if the given version is newer (more recent)
        """
        if not any([self.version, version_to_compare]):
            logger.error("Object version is not set, so the comparizon is impossible.")
            return None

        # if a plugin is given
        if isinstance(version_to_compare, QgisPlugin):
            # take the opportunity to check the name is the same
            if not self.name == version_to_compare.name:
                logger.warning(
                    "Be careful, the plugin to compare seems to be different: "
                    f"{self.name} != {version_to_compare.name}"
                )
            # store the version string
            version_to_compare = version_to_compare.version

        # load object version as packaging.Version object
        try:
            plugin_version = Version(self.version)
        except InvalidVersion as err:
            logger.error(
                f"Plugin {self.name} (current) uses an incompatible versioning scheme: "
                f"{self.version}. It's not Semver (even prefixed by 'v'), nor Calver or "
                "any of supported specification. See https://peps.python.org/pep-0440/. "
                f"Trace: {err}"
            )
            return None

        # load version to compare as packaging.Version object
        try:
            version_to_compare = Version(version_to_compare)
        except InvalidVersion as err:
            logger.error(
                f"Plugin {self.name} (to compare) uses an incompatible versioning scheme: "
                f"{self.version}. It's not Semver (even prefixed by 'v'), nor Calver or "
                "any of supported specification. See https://peps.python.org/pep-0440/. "
                f"Trace: {err}"
            )
            return None

        logger.debug(f"Comparing versions: {plugin_version} and {version_to_compare}")

        return plugin_version < version_to_compare

    @property
    def uri_to_zip(self) -> str:
        """Get the plugin URI. Can be either a local path or remote URL (in that case
            it's mapped to dowload_url property).

        Returns:
            str: URI (path or URL) to the plugin ZIP archive
        """
        if (
            self.url
            and isinstance(self.url, str)
            and self.url.startswith(("http://", "https://"))
        ):
            return self.download_url
        elif (
            self.url
            and isinstance(self.url, str)
            and (self.url.startswith("file://") or Path(self.url))
        ):
            uri_or_path = self.url
            if self.url.startswith("file://"):
                uri_or_path = self.url[7:]
                logger.debug(
                    f"URI cleaning: 'file://' protocol prefix removed. Result: {uri_or_path}"
                )
            return Path(expandvars(expanduser(uri_or_path)))
        else:
            return self.url


# #############################################################################
# ##### Stand alone program ########
# ##################################

if __name__ == "__main__":
    """Standalone execution."""
    sample_plugin_complete = {
        "name": "french_locator_filter",
        "version": "1.0.4",
        "url": "https://plugins.qgis.org/plugins/french_locator_filter/version/1.0.4/download/",
        "location": "remote",
    }

    plugin_obj_one: QgisPlugin = QgisPlugin.from_dict(sample_plugin_complete)
    assert plugin_obj_one.url == plugin_obj_one.download_url
    print(plugin_obj_one)

    sample_plugin_minimal = {
        "name": "french_locator_filter",
        "version": "1.0.4",
        "official_repository": True,
    }

    plugin_obj_two = QgisPlugin.from_dict(sample_plugin_minimal)
    print(plugin_obj_two)

    assert plugin_obj_one == plugin_obj_two

    sample_plugin_unofficial = {
        "name": "Geotuileur",
        "version": "1.0.0",
        "official_repository": False,
        "repository_url_xml": "https://oslandia.gitlab.io/qgis/ign-geotuileur/plugins.xml",
    }

    plugin_obj_three: QgisPlugin = QgisPlugin.from_dict(sample_plugin_unofficial)
    print(plugin_obj_three)
    print(plugin_obj_three.folder_name)
    print(plugin_obj_three.installation_folder_name)
    print(plugin_obj_three.download_url)

    sample_plugin_different_name = {
        "name": "Layers menu from project",
        "version": "2.1.0",
        "url": "https://plugins.qgis.org/plugins/menu_from_project/version/2.1.0/download/",
        "location": "remote",
        "plugin_id": 1846,
    }

    plugin_obj_four: QgisPlugin = QgisPlugin.from_dict(sample_plugin_different_name)
    print(plugin_obj_four.url)
    print(plugin_obj_four.official_repository)
    print(plugin_obj_four.folder_name)
    print(plugin_obj_four.installation_folder_name)
    print(plugin_obj_four.folder_name)
    assert plugin_obj_four.url == plugin_obj_four.download_url

    sample_zip = (
        Path.home() / ".cache/qgis-deployment-toolbelt/plugins/qompligis_v1-1-0.zip"
    )
    plugin_from_zip: QgisPlugin = QgisPlugin.from_zip(sample_zip)
    print(plugin_from_zip)

    sample_plugin_qtribu = {
        "name": "QTribu",
        "version": "0.14.2",
        "official_repository": True,
        "folder_name": "qtribu",
    }
    plugin_obj_five: QgisPlugin = QgisPlugin.from_dict(sample_plugin_qtribu)
    print(plugin_obj_five.url)

    sample_plugin_qtribu_newer: QgisPlugin = QgisPlugin.from_dict(
        {
            "name": "QTribu",
            "version": "0.15.0",
            "official_repository": True,
            "folder_name": "qtribu",
        }
    )

    assert plugin_obj_five.is_older_than(sample_plugin_qtribu_newer) is True
    assert sample_plugin_qtribu_newer.is_older_than(plugin_obj_five) is False

    plugin_dict_local = {
        "name": "STSI Locator Filter",
        "folder_name": "stsi_locator_filter",
        "official_repository": False,
        "plugin_id": 202303,
        "location": "local",
        "url": "~/Git/Oslandia/QGIS/stsi-plugin-qgis-geocoder-locator-filter/stsi_locator_filter.1.0.0.zip",
        "version": "1.0.0",
    }
    plugin_obj_six = QgisPlugin.from_dict(plugin_dict_local)
    print(plugin_obj_six.uri_to_zip)



#! python3  # noqa: E265

"""
    Utilities specific for Windows.

    Author: Julien Moura (https://github.com/guts)

    Inspired from py-setenv: <https://github.com/beliaev-maksim/py_setenv> (MIT)
"""

# #############################################################################
# ########## Libraries #############
# ##################################


# Standard library
import ctypes
import logging
from enum import Enum
from os import sep  # required since pathlib strips trailing whitespace
from pathlib import Path
from sys import platform as opersys

# Imports depending on operating system
if opersys == "win32":
    """windows"""
    # standard
    import winreg

    # 3rd party
    import win32gui


# #############################################################################
# ########## Globals ###############
# ##################################

logger = logging.getLogger(__name__)

if opersys == "win32":
    """windows"""
    system_hkey = (
        winreg.HKEY_LOCAL_MACHINE,
        r"SYSTEM\CurrentControlSet\Control\Session Manager\Environment",
    )
    user_hkey = (winreg.HKEY_CURRENT_USER, r"Environment")


# #############################################################################
# ########## Classes ###############
# ##################################


class ExtendedNameFormat(Enum):
    """Possible values for user name in a Windows Active Directory context.
    See references:

    - https://learn.microsoft.com/windows/win32/api/secext/ne-secext-extended_name_format
    - https://mhammond.github.io/pywin32/win32api__GetUserNameEx_meth.html
    - https://github.com/mhammond/pywin32/blob/dde12b8ef274a157aede18f46cae5b44f112be17/win32/Lib/win32con.py#L4963-L4973

    """

    # An unknown name type.
    NameUnknown = 0
    # The fully qualified distinguished name (for example, CN=Jeff Smith,OU=Users,DC=Engineering,DC=Microsoft,DC=Com).
    NameFullyQualifiedDN = 1
    # A legacy account name (for example, Engineering\JSmith).
    # The domain-only version includes trailing backslashes (\).
    NameSamCompatible = 2
    # A "friendly" display name (for example, Jeff Smith).
    # The display name is not necessarily the defining relative distinguished name (RDN).
    NameDisplay = 3
    # A GUID string that the IIDFromString function returns (for example,
    # {4fa050f0-f561-11cf-bdd9-00aa003a77b6}).
    NameUniqueId = 6
    # The complete canonical name (for example,
    # engineering.microsoft.com/software/someone).
    # The domain-only version includes a trailing forward slash (/).
    NameCanonical = 7
    # The user principal name (for example, someone@example.com).
    NameUserPrincipal = 8
    # The same as NameCanonical except that the rightmost forward slash (/) is replaced
    # with a new line character (\n), even in a domain-only case (for example,
    # engineering.microsoft.com/software\nJSmith).
    NameCanonicalEx = 9
    # The generalized service principal name (for example,
    # www/www.microsoft.com@microsoft.com).
    NameServicePrincipal = 10
    # The DNS domain name followed by a backward-slash and the SAM user name.
    NameDnsDomain = 12
    # The first name or given name of the user. Note: This type is only available for
    # GetUserNameEx calls for an Active Directory user.
    NameGivenName = 13
    # The last name or surname of the user. Note: This type is only available for
    # GetUserNameEx calls for an Active Directory user.
    NameSurname = 14


# #############################################################################
# ########## Functions #############
# ##################################


def delete_environment_variable(envvar_name: str, scope: str = "user") -> bool:
    """Deletes environment variable.

    Args:
        envvar_name (str): environment variable name (= key) to delete
        scope (str, optional): environment variable scope. Must be "user" or "system",
            defaults to "user". Defaults to "user".

    Returns:
        bool: True is the variable has been successfully deleted
    """
    # user or system
    if scope == "user":
        hkey = user_hkey
    else:
        hkey = system_hkey

    # get it to check if variable exits
    try:
        get_environment_variable(envvar_name=envvar_name, scope=scope)
    except Exception:
        return False

    # try to delete the variable
    try:
        with winreg.OpenKey(*hkey, access=winreg.KEY_ALL_ACCESS) as key:
            winreg.DeleteValue(key, envvar_name)
            return True
    except OSError as err:
        logger.error(
            f"Delete variable '{envvar_name}' from scope '{scope}' failed. Trace: {err}"
        )
        return False


def get_environment_variable(envvar_name: str, scope: str = "user") -> str | None:
    """Get environment variable from Windows registry.

    Args:
        envvar_name (str): environment variable name (= key) to retrieve
        scope (str, optional): environment variable scope. Must be "user" or "system",
            defaults to "user". Defaults to "user".

    Returns:
        Optional[str]: environment variable value or None if not found
    """
    # user or system
    if scope == "user":
        hkey = user_hkey
    else:
        hkey = system_hkey

    # try to get the value
    try:
        with winreg.OpenKey(*hkey, access=winreg.KEY_READ) as key:
            value, _ = winreg.QueryValueEx(key, envvar_name)
        return value
    except OSError:
        logger.error(
            f"Environment variable {envvar_name} not found in registry (scope: {scope}"
        )
        return None


def get_current_user_extended_data(extended_name_format: ExtendedNameFormat) -> str:
    """Get current user full data extended with Active Directory informations.

    This method uses the ctypes module since the upper-level method implemented in
    `pywin32.win32api` (`win32api.GetUserNameEx(extended_name_format.value)`) raises an
    error when the specified format is not reachable.

    Inspired from: https://stackoverflow.com/a/70182936/2556577

    Returns:
        str: use data in the specified format

    Example:

        .. code-block:: python

            from qgis_deployment_toolbelt.utils.win32utils import (
                ExtendedNameFormat,
                get_current_user_extended_data,
            )

            user_data = {
                k.name: get_current_user_extended_data(k)
                for k in ExtendedNameFormat
            }

            print(user_data)

    """
    format_index = extended_name_format.value

    # use system DLL to call API
    GetUserNameEx = ctypes.windll.secur32.GetUserNameExW

    size = ctypes.pointer(ctypes.c_ulong(0))
    GetUserNameEx(format_index, None, size)

    nameBuffer = ctypes.create_unicode_buffer(size.contents.value)
    GetUserNameEx(format_index, nameBuffer, size)

    return nameBuffer.value


def normalize_path(input_path: Path, add_trailing_slash_if_dir: bool = True) -> str:
    r"""Returns a path as normalized and fully escaped for Windows old-school file style.

    :param Path input_path: path to normalize
    :param bool add_trailing_slash_if_dir: add a trailing slash if the input is a folder,\
        defaults to True

    :return str: normalized path as string

    :example:

    .. code-block:: python

        t = Path(r'C:\Users\risor\Documents\GitHub\Geotribu\qtribu\qtribu\resources\images')
        print(normalize_path(t))
        > C:\\Users\\risor\\Documents\\GitHub\\Geotribu\\qtribu\\qtribu\\resources\\images\\

    """
    if input_path.is_dir() and add_trailing_slash_if_dir:
        return repr(str(input_path.resolve()) + sep).replace("'", "")
    else:
        return repr(str(input_path.resolve())).replace("'", "")


def refresh_environment() -> bool:
    """This ensures that changes to Windows registry are immediately propagated.
    Useful to refresh after have updated the environment variables.

    A method by Geoffrey Faivre-Malloy and Ronny Lipshitz.
    Source: https://gist.github.com/apetrone/5937002

    Returns:
        bool: True if the environment has been refreshed
    """
    # broadcast settings change
    HWND_BROADCAST: int = 0xFFFF
    WM_SETTINGCHANGE: int = 0x001A
    SMTO_ABORTIFHUNG: int = 0x0002
    send_parameter = "Environment"

    res1 = res2 = None
    try:
        res1, res2 = win32gui.SendMessageTimeout(
            HWND_BROADCAST, WM_SETTINGCHANGE, 0, send_parameter, SMTO_ABORTIFHUNG, 100
        )
    except NameError:
        logger.critical(" name 'win32gui' is not defined")
    if not res1:
        logger.warning(
            f"Refresh environment failed: {bool(res1)}, {res2}, from SendMessageTimeout"
        )
        return False
    else:
        return True


def set_environment_variable(
    envvar_name: str, envvar_value: str, scope: str = "user"
) -> bool:
    """Creates/replaces environment variable.

    Args:
        envvar_name (str): name (= key) of environment variable to set or replace.
        envvar_value (str): value to set for the environment variable
        scope (str, optional): environment variable scope. Must be "user" or "system",
            defaults to "user". Defaults to "user".

    Returns:
        bool: True is the variable has been successfully set
    """
    # user or system
    if scope == "user":
        hkey = user_hkey
    else:
        hkey = system_hkey

    # try to set the value
    try:
        with winreg.OpenKey(*hkey, access=winreg.KEY_WRITE) as key:
            winreg.SetValueEx(key, envvar_name, 0, winreg.REG_SZ, envvar_value)
        return True
    except OSError as err:
        logger.error(
            f"Set variable '{envvar_name}' with value '{envvar_value}' to "
            f"scope '{scope}' failed. Trace: {err}"
        )
        return False


# #############################################################################
# ##### Stand alone program ########
# ##################################

if __name__ == "__main__":
    """Standalone execution."""
    # pass
    t = Path("C:/Users/risor/Documents/GitHub/Geotribu/qtribu/qtribu/resources/images")
    print(normalize_path(t))



#! python3  # noqa E265

"""
    Usage from the repo root folder:

    .. code-block:: bash
        # for whole tests
        python -m unittest tests.test_utils_file_downloader
        # for specific test
        python -m unittest tests.test_utils_file_downloader.TestUtilsFileDownloader.test_download_file_exists
"""

# standard library
import tempfile
import unittest
from pathlib import Path

# 3rd party
from requests.exceptions import ConnectionError, HTTPError

# project
from qgis_deployment_toolbelt.utils.file_downloader import download_remote_file_to_local

# ############################################################################
# ########## Classes #############
# ################################


class TestUtilsFileDownloader(unittest.TestCase):
    """Test package utilities."""

    def test_download_file_exists(self):
        """Test download remote file locally to a file which already exists."""
        with tempfile.TemporaryDirectory(
            prefix="qdt_test_downloader_", ignore_cleanup_errors=True
        ) as tmpdirname:
            # file that already exist locally
            downloaded_file = download_remote_file_to_local(
                remote_url_to_download="https://raw.githubusercontent.com/qgis-deployment/qgis-deployment-toolbelt-cli/main/README.md",
                local_file_path=Path(tmpdirname).joinpath("README_from_remote.md"),
            )
            self.assertIsInstance(downloaded_file, Path)
            self.assertTrue(downloaded_file.exists())
            self.assertTrue(downloaded_file.is_file())

        # disabling stream mode
        with tempfile.TemporaryDirectory(
            prefix="qdt_test_downloader_nostream_", ignore_cleanup_errors=True
        ) as tmpdirname:
            # file that already exist locally
            downloaded_file = download_remote_file_to_local(
                remote_url_to_download="https://raw.githubusercontent.com/qgis-deployment/qgis-deployment-toolbelt-cli/main/README.md",
                local_file_path=Path(tmpdirname).joinpath("README_from_remote.md"),
                use_stream=False,
            )
            self.assertIsInstance(downloaded_file, Path)
            self.assertTrue(downloaded_file.exists())
            self.assertTrue(downloaded_file.is_file())

    def test_download_file_raise_http_error(self):
        """Test download handling an HTTP error."""
        with tempfile.TemporaryDirectory(
            prefix="qdt_test_downloader_", ignore_cleanup_errors=True
        ) as tmpdirname:
            target_path = Path(tmpdirname).joinpath("dl_404.html")
            with self.assertRaises(HTTPError):
                download_remote_file_to_local(
                    remote_url_to_download="https://qgis.org/fake-page",
                    local_file_path=target_path,
                )
            self.assertFalse(target_path.exists())
            self.assertFalse(target_path.is_file())

    def test_download_file_raise_url_error(self):
        """Test download with a bad URL."""
        with tempfile.TemporaryDirectory(
            prefix="qdt_test_downloader_", ignore_cleanup_errors=True
        ) as tmpdirname:
            target_path = Path(tmpdirname).joinpath("fake_file.dmg")
            with self.assertRaises(ConnectionError):
                download_remote_file_to_local(
                    remote_url_to_download="https://fake_url/youpi.dmg",
                    local_file_path=target_path,
                )
            self.assertFalse(target_path.exists())
            self.assertFalse(target_path.is_file())


# ############################################################################
# ####### Stand-alone run ########
# ################################
if __name__ == "__main__":
    unittest.main()



#! python3  # noqa: E265

"""
    QDT QGIS Profile object and related sugar.

    Author: Julien Moura (https://github.com/guts)
"""


# #############################################################################
# ########## Libraries #############
# ##################################

# special
from __future__ import annotations

# standard
import json
import logging
import tempfile
from pathlib import Path
from shutil import copy2, copytree
from typing import Literal

# 3rd party
from packaging.version import InvalidVersion, Version

# Package
from qgis_deployment_toolbelt.constants import (
    OSConfiguration,
    get_qdt_working_directory,
)
from qgis_deployment_toolbelt.plugins.plugin import QgisPlugin
from qgis_deployment_toolbelt.profiles.qgis_ini_handler import QgisIniHelper
from qgis_deployment_toolbelt.utils.check_path import check_path

# #############################################################################
# ########## Globals ###############
# ##################################

# logs
logger = logging.getLogger(__name__)


# #############################################################################
# ########## Classes ###############
# ##################################
class QdtProfile:
    """Object definition for QGIS Profile handled by QDT."""

    # optional mapping on attributes names.
    # {attribute_name_in_output_object: attribute_name_from_input_file}  # noqa: E800
    ATTR_MAP = {
        "qgis_maximum_version": "qgisMaximumVersion",
        "qgis_minimum_version": "qgisMinimumVersion",
    }

    def __init__(
        self,
        alias: str | None = None,
        author: str | None = None,
        description: str | None = None,
        email: str | None = None,
        folder: Path | None = None,
        icon: str | None = None,
        json_ref_path: Path | None = None,
        loaded_from_json: bool = False,
        name: str | None = None,
        plugins: list | None = None,
        qgis_maximum_version: str | None = None,
        qgis_minimum_version: str | None = None,
        rules: list[dict] | None = None,
        splash: str | None = None,
        version: str | None = None,
        **kwargs,
    ):
        """Initialize a QDT Profile object.

        :param str name: name of the shortcut that will be created
        :param str author: profile author name
        """
        # store QDT working folder
        self.qdt_working_folder = get_qdt_working_directory()
        # retrieve operating system specific configuration
        self.os_config = OSConfiguration.from_opersys()

        # default values for immutable attributes
        self.loaded_from_json = loaded_from_json

        # default values for attributes/properties that can be get/set
        self._alias = None
        self._author = None
        self._description = None
        self._email = None
        self._folder = None
        self._icon = None
        self._json_ref_path = None
        self._name = None
        self._splash = None
        self._plugins = None
        self._qgis_maximum_version = None
        self._qgis_minimum_version = None
        self._rules = None
        self._version = None

        # if values have been passed, so use them as objects attributes.
        # attributes are prefixed by an underscore '_'
        if alias:
            self._alias = alias
        if author:
            self._author = author
        if description:
            self._description = description
        if email:
            self._email = email
        if folder:
            self._folder = folder
        if icon:
            self._icon = icon
        if json_ref_path:
            self._json_ref_path = json_ref_path
        if name:
            self._name = name
        if plugins:
            self._plugins = plugins
        if qgis_maximum_version:
            self._qgis_maximum_version = qgis_maximum_version
        if qgis_minimum_version:
            self._qgis_minimum_version = qgis_minimum_version
        if rules:
            self._rules = rules
        if splash:
            self._splash = splash
        if version:
            self._version = version

    @classmethod
    def from_json(
        cls, profile_json_path: Path, profile_folder: Path | None = None
    ) -> QdtProfile:
        """Load profile from a profile.json file.

        Args:
            profile_json_path (Path): path to the profile json file
            profile_folder (Path | None, optional): path to the profile folder,
                defaults to None.

        Returns:
            QdtProfile: QdtProfile object with attributes filled from JSON.

        Example:

            .. code-block:: python

                QdtProfile = QdtProfile.from_json(
                        Path(src_profile / "profile.json")
                    )
                print(profile.splash.resolve())
        """
        # checks
        check_path(
            input_path=profile_json_path,
            must_be_a_file=True,
            must_exists=True,
            must_be_readable=True,
        )
        if profile_folder:
            check_path(
                input_path=profile_folder,
                must_be_a_folder=True,
                must_exists=True,
                must_be_readable=True,
            )

        # load JSON
        with profile_json_path.open(mode="r", encoding="utf8") as in_profile_json:
            profile_data = json.load(in_profile_json)

        # map attributes names
        for k, v in cls.ATTR_MAP.items():
            profile_data[k] = profile_data.pop(v, None)

        # return new instance with loaded object
        return cls(
            folder=profile_folder,
            json_ref_path=profile_json_path,
            loaded_from_json=True,
            **profile_data,
        )

    @property
    def alias(self) -> str:
        """Returns the profile's alias.

        :return str: profile alias
        """
        return self._alias

    @property
    def folder(self) -> Path:
        """Returns the path to the folder where the profile is stored.

        Returns:
            Path: profile folder path
        """
        if isinstance(self._folder, Path):
            return self._folder.resolve()
        else:
            return self._folder

    @property
    def is_loaded_from_json(self) -> bool:
        """Tells if the profile has been loaded from a JSON file.

        :return bool: True if the profile has been loaded from a JSON file
        """
        return self.loaded_from_json

    @property
    def icon(self) -> str | None:
        """Returns the icon as specified into the original profile.json.

        :return str: profile icon value
        """
        if self._icon:
            return self._icon

    @property
    def icon_path(self) -> Path:
        """Returns the path to the profile's icon into the profile folder.

        :return Path: profile icon path
        """
        if self._folder and self._icon:
            return self._folder.joinpath(self._icon)

    @property
    def json_ref_path(self) -> Path:
        """Returns the path to the corresponding JSON path.

        :return Path: profile json_ref_path path
        """
        return self._json_ref_path.resolve()

    @property
    def name(self) -> str | None:
        """Returns the profile's name. If not set, the folder name is used.

        Returns:
            str: profile's name
        """
        if self._name:
            return self._name
        elif self._folder:
            return self._folder
        else:
            return None

    @property
    def path_in_qgis(self) -> Path:
        """Returns the path to the folder where the profile is stored in QGIS 3
            (= installed).

        Returns:
            Path: path to the installed (i.e. in QGIS) profile folder
        """
        return self.os_config.qgis_profiles_path.joinpath(self.name)

    @property
    def plugins(self) -> list[QgisPlugin]:
        """Returns the plugins associated with the profile.

        Returns:
            List[QgisPlugin]: list of plugins
        """
        if self._plugins:
            return [QgisPlugin.from_dict(p) for p in self._plugins]
        else:
            return []

    @property
    def rules(self) -> list[dict] | None:
        """Returns the rules associated with the profile.

        Returns:
            list[rules]: list of rules
        """
        return self._rules

    @property
    def splash(self) -> str | Path | None:
        """Returns the profile splash image as path if can be resolved or as string.

        Returns:
            Union[str, Path]: path to the profile splash image
        """
        if (
            self._splash
            and isinstance(self.folder, Path)
            and self.folder.joinpath(self._splash).is_file()
        ):
            return self.folder.joinpath(self._splash)
        else:
            return self._splash

    @property
    def version(self) -> str | None:
        """Returns the profile version as string.

        Returns:
            str: version
        """
        return self._version

    def is_older_than(self, version_to_compare: str | QdtProfile) -> bool | None:
        """Determine if the actual object version is older than the given version to
            compare.

        Args:
            version_to_compare (Union[str, QdtProfile]): given version to compare with object
            version

        Returns:
            bool: True if the given version is newer (more recent)
        """
        if not any([self.version, version_to_compare]):
            logger.error("Object version is not set, so the comparizon is impossible.")
            return None

        # if a profile is given
        if isinstance(version_to_compare, QdtProfile):
            # take the opportunity to check the name is the same
            if not self.name == version_to_compare.name:
                logger.warning(
                    "Be careful, the profile to compare seems to be different: "
                    f"{self.name} != {version_to_compare.name}"
                )
            # store the version string
            version_to_compare = version_to_compare.version

        # load object version as packaging.Version object
        try:
            profile_version = Version(self.version)
        except InvalidVersion as err:
            logger.error(
                f"Profile {self.name} uses an incompatible versioning scheme: {self.version}."
                "It's not Semver (even prefixed by 'v'), nor Calver or any of "
                "supported specification. See https://peps.python.org/pep-0440/. "
                f"Trace: {err}"
            )
            return None

        # load version to compare as packaging.Version object
        try:
            version_to_compare = Version(version_to_compare)
        except InvalidVersion as err:
            logger.error(
                f"Version to compare uses an incompatible versioning scheme: {self.version}."
                "It's not Semver (even prefixed by 'v'), nor Calver or any of "
                "supported specification. See https://peps.python.org/pep-0440/. "
                f"Trace: {err}"
            )
            return None

        logger.debug(f"Comparing versions: {profile_version} and {version_to_compare}")

        return profile_version < version_to_compare

    def status(self) -> Literal["downloaded", "installed", "unknown"]:
        """Determine current profile status: downloaded (in QDT working folder),
        installed (in QGIS3/profiles) or unknown.

        Returns:
            str: one of "downloaded", "installed", "unknown"
        """
        if not isinstance(self.folder, Path):
            return "unknown"

        if self.folder.is_relative_to(self.os_config.qgis_profiles_path):
            return "installed"
        elif self.folder.is_relative_to(self.qdt_working_folder):
            return "downloaded"
        else:
            return "unknown"

    @property
    def installed_profile(self) -> QdtProfile | None:
        """Returns the installed profile object only if the corresponding profiles.json
            exists.

        Returns:
            QdtProfile | None: QdtProfile object if profile.json exists, else None
        """

        if self.status() == "installed":
            return self
        elif self.status() == "downloaded":
            if self.path_in_qgis.joinpath("profile.json").is_file():
                return self.from_json(
                    profile_json_path=self.path_in_qgis.joinpath("profile.json"),
                    profile_folder=self.path_in_qgis.resolve(),
                )
            else:
                # TODO: develop a class method to load a profile from a folder
                pass
        return None

    # -- QGIS*.ini files --
    def has_qgis3_ini_file(self) -> bool:
        """Determine if a QGIS/QGIS3.ini file exists in the profile folder.

        Returns:
            bool: True if a QGIS/QGIS3.ini file exists in the profile folder.
        """
        return check_path(
            input_path=self.folder.joinpath("QGIS/QGIS3.ini"),
            must_be_a_file=True,
            must_be_a_folder=False,
            must_be_readable=True,
            must_exists=True,
            raise_error=False,
        )

    def has_qgis3customization_ini_file(self) -> bool:
        """Determine if a QGIS/QGISCUSTOMIZATION3.ini file exists in the profile folder.

        Returns:
            bool: True if a QGIS/QGISCUSTOMIZATION3.ini file exists in the profile folder.
        """
        return check_path(
            input_path=self.folder.joinpath("QGIS/QGISCUSTOMIZATION3.ini"),
            must_be_a_file=True,
            must_be_a_folder=False,
            must_be_readable=True,
            must_exists=True,
            raise_error=False,
        )

    def get_qgis3ini_helper(self) -> QgisIniHelper:
        """Return the QGIS3 ini helper for the profile configuration.

        Returns:
            QgisIniHelper: Ini helper loaded with profile's QGIS/QGIS3.ini
        """
        logger.debug(
            f"Returning QGISCUSTOMIZATION3.ini helper for profile '{self.name}' using "
            f"this file: {self.folder.joinpath('QGIS/QGISCUSTOMIZATION3.ini')}"
        )
        return QgisIniHelper(
            ini_filepath=self.folder.joinpath("QGIS/QGIS3.ini"),
            ini_type="profile_qgis3",
        )

    def get_qgis3customizationini_helper(self) -> QgisIniHelper:
        """Return the QGIS3 ini helper for the profile customization.

        Returns:
            QgisIniHelper: Ini helper loaded with profile's QGIS/QGISCUSTOMIZATION3.ini
        """
        logger.debug(
            f"Returning QGISCUSTOMIZATION3.ini helper for profile '{self.name}' using "
            f"this file: {self.folder.joinpath('QGIS/QGISCUSTOMIZATION3.ini')}"
        )
        return QgisIniHelper(
            ini_filepath=self.folder.joinpath("QGIS/QGISCUSTOMIZATION3.ini"),
            ini_type="profile_qgis3customization",
        )

    def merge_to(self, dst: QdtProfile) -> None:
        """Merge QdtProfile to another profile


        Args:
            dst (QdtProfile): _description_
        """
        with tempfile.TemporaryDirectory(
            prefix=f"QDT_merge_profile_{self.name}_", ignore_cleanup_errors=True
        ) as tmpdirname:
            logger.info(
                f"Merge profile {self.name} with {tmpdirname} temporary directory"
            )
            # Copy source QdtProfile folder
            copytree(
                self.folder,
                tmpdirname,
                copy_function=copy2,
                dirs_exist_ok=True,
            )
            # Merge INI files
            tmp_profile = QdtProfile(folder=Path(tmpdirname))

            # QGIS3
            if self.has_qgis3_ini_file() and dst.has_qgis3_ini_file():
                # Copy current installed file
                copy2(
                    dst.get_qgis3ini_helper().ini_filepath,
                    tmp_profile.get_qgis3ini_helper().ini_filepath,
                )
                # Merge
                self.get_qgis3ini_helper().merge_to(tmp_profile.get_qgis3ini_helper())

            # QGISCUSTOMIZATION3
            if (
                self.has_qgis3customization_ini_file()
                and dst.has_qgis3customization_ini_file()
            ):
                # Copy current installed file
                copy2(
                    dst.get_qgis3customizationini_helper().ini_filepath,
                    tmp_profile.get_qgis3customizationini_helper().ini_filepath,
                )
                # Merge
                self.get_qgis3customizationini_helper().merge_to(
                    tmp_profile.get_qgis3customizationini_helper()
                )

            logger.info(f"Copying {tmpdirname} to {dst.path_in_qgis}")
            copytree(
                tmpdirname,
                dst.path_in_qgis,
                copy_function=copy2,
                dirs_exist_ok=True,
            )


# #############################################################################
# ##### Stand alone program ########
# ##################################

if __name__ == "__main__":
    """Standalone execution."""
    pass



#! python3  # noqa: E265

"""
    Test CLI's rules context export command.

    Author: Julien Moura (Oslandia)
"""

# #############################################################################
# ########## Libraries #############
# ##################################


# 3rd party
import pytest

# project
from qgis_deployment_toolbelt import cli

# #############################################################################
# ######## Classes #################
# ##################################


@pytest.mark.parametrize("option", ("-h", "--help"))
def test_cli_export_rules_context_help(capsys, option):
    """Test CLI help."""
    with pytest.raises(SystemExit):
        cli.main(["export-rules-context", option])

    _, err = capsys.readouterr()

    assert err == ""


def test_cli_export_rules_context(capsys):
    """Test CLI."""
    with pytest.raises(SystemExit):
        cli.main(["export-rules-context"])

    _, err = capsys.readouterr()

    assert err == ""



#! python3  # noqa E265

"""Usage from the repo root folder:

    .. code-block:: python

        # for whole test
        python -m unittest tests.test_job_environment_variables
        # for specific
        python -m unittest tests.test_job_environment_variables
            .TestJobEnvironmentVariables.test_environment_variables_set
"""

# #############################################################################
# ########## Libraries #############
# ##################################


# Standard library
import unittest
from os.path import expanduser
from pathlib import Path
from sys import platform as opersys

# package
from qgis_deployment_toolbelt.jobs.job_environment_variables import (
    JobEnvironmentVariables,
)
from qgis_deployment_toolbelt.utils.str2bool import str2bool

# 3rd party


# conditional imports
if opersys == "win32":
    from qgis_deployment_toolbelt.utils.win32utils import get_environment_variable
elif opersys == "linux":
    from qgis_deployment_toolbelt.utils.linux_utils import get_environment_variable


# #############################################################################
# ########## Classes ###############
# ##################################


class TestJobEnvironmentVariables(unittest.TestCase):
    """Test module."""

    # -- Standard methods --------------------------------------------------------
    @classmethod
    def setUpClass(cls):
        """Executed when module is loaded before any test."""
        cls.good_scenario_files = sorted(
            Path("tests/fixtures/").glob("scenarios/good_*.y*ml")
        )

    # standard methods
    def setUp(self):
        """Fixtures prepared before each test."""
        pass

    def tearDown(self):
        """Executed after each test."""
        pass

    # -- TESTS ---------------------------------------------------------
    @unittest.skipIf(opersys == "darwin", f"Not supported operating system: {opersys}.")
    def test_environment_variables_set_unset(self):
        """Test YAML loader"""
        fake_env_vars = [
            {
                "action": "add",
                "name": "QDT_TEST_ENV_VAR",
                "scope": "user",
                "value": "this is a custom value",
            },
            {
                "action": "add",
                "name": "QDT_TEST_URL_API_PLUGIN",
                "scope": "user",
                "value": "http://api.intra.net",
                "value_type": "url",
            },
            {
                "action": "add",
                "name": "QDT_TEST_FAKE_ENV_VAR_BOOL",
                "scope": "user",
                "value": True,
                "value_type": "bool",
            },
            {
                "action": "add",
                "name": "QDT_TEST_FAKE_ENV_VAR_PATH",
                "scope": "user",
                "value": "~/scripts/qgis_startup.py",
                "value_type": "path",
            },
        ]
        job_env_vars = JobEnvironmentVariables(fake_env_vars)
        job_env_vars.run()

        # check if environment variables have been set
        if opersys == "win32":
            self.assertTrue(
                str2bool(
                    get_environment_variable("QDT_TEST_FAKE_ENV_VAR_BOOL", "user")
                ),
                True,
            )
            self.assertEqual(
                get_environment_variable("QDT_TEST_FAKE_ENV_VAR_PATH"),
                str(Path(expanduser("~/scripts/qgis_startup.py")).resolve()),
            )
            self.assertEqual(
                get_environment_variable("QDT_TEST_URL_API_PLUGIN"),
                "http://api.intra.net",
            )
            self.assertEqual(
                get_environment_variable("QDT_TEST_ENV_VAR"),
                "this is a custom value",
            )

            # clean up
            fake_env_vars = [
                {
                    "name": "QDT_TEST_ENV_VAR",
                    "scope": "user",
                    "action": "remove",
                },
                {
                    "name": "QDT_TEST_URL_API_PLUGIN",
                    "scope": "user",
                    "action": "remove",
                },
                {
                    "name": "QDT_TEST_FAKE_ENV_VAR_BOOL",
                    "scope": "user",
                    "action": "remove",
                },
                {
                    "name": "QDT_TEST_FAKE_ENV_VAR_PATH",
                    "scope": "user",
                    "action": "remove",
                },
            ]
            job_env_vars = JobEnvironmentVariables(fake_env_vars)
            job_env_vars.run()

            self.assertIsNone(
                get_environment_variable("QDT_TEST_FAKE_ENV_VAR_BOOL", "user")
            )
            self.assertIsNone(
                get_environment_variable("QDT_TEST_FAKE_ENV_VAR_PATH"),
            )
        elif opersys == "linux":
            self.assertTrue(
                str2bool(
                    get_environment_variable("QDT_TEST_FAKE_ENV_VAR_BOOL", "user")
                ),
                True,
            )
            self.assertEqual(
                get_environment_variable("QDT_TEST_FAKE_ENV_VAR_PATH"),
                str(Path(expanduser("~/scripts/qgis_startup.py")).resolve()),
            )
            self.assertEqual(
                get_environment_variable("QDT_TEST_URL_API_PLUGIN"),
                "http://api.intra.net",
            )
            self.assertEqual(
                get_environment_variable("QDT_TEST_ENV_VAR"),
                "this is a custom value",
            )

            # clean up
            fake_env_vars = [
                {
                    "name": "QDT_TEST_ENV_VAR",
                    "scope": "user",
                    "action": "remove",
                },
                {
                    "name": "QDT_TEST_URL_API_PLUGIN",
                    "scope": "user",
                    "action": "remove",
                },
                {
                    "name": "QDT_TEST_FAKE_ENV_VAR_BOOL",
                    "scope": "user",
                    "action": "remove",
                },
                {
                    "name": "QDT_TEST_FAKE_ENV_VAR_PATH",
                    "scope": "user",
                    "action": "remove",
                },
            ]
            job_env_vars = JobEnvironmentVariables(fake_env_vars)
            job_env_vars.run()

            self.assertIsNone(
                get_environment_variable("QDT_TEST_FAKE_ENV_VAR_BOOL", "user")
            )
            self.assertIsNone(
                get_environment_variable("QDT_TEST_FAKE_ENV_VAR_PATH"),
            )

    def test_prepare_value(self):
        """Test prepare_value method"""
        job_env_vars = JobEnvironmentVariables([])
        value_test = f"tests/{Path(__file__).name}"
        self.assertEqual(
            job_env_vars.prepare_value(value=value_test, value_type="path"),
            str(Path().resolve() / value_test),
        )
        value_test = "imaginary/path"
        if opersys == "win32":
            self.assertEqual(
                job_env_vars.prepare_value(value=value_test, value_type="path"),
                str(Path().resolve() / value_test),
            )
            self.assertEqual(
                job_env_vars.prepare_value(value=[], value_type="str"),
                "[]",
            )
        else:
            self.assertEqual(
                job_env_vars.prepare_value(value=value_test, value_type="path"),
                str(Path().resolve() / value_test),
            )
            self.assertEqual(
                job_env_vars.prepare_value(value=[]),
                "[]",
            )



#! python3  # noqa: E265

"""Reader for qdt-files.json generated with tree."""

# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import logging
from typing import TypedDict

# #############################################################################
# ########## Globals ###############
# ##################################

logger = logging.getLogger(__name__)

# #############################################################################
# ########## Classes ###############
# ##################################


class Treeitem(TypedDict):
    type: str
    name: str
    contents: list[dict] | None


# #############################################################################
# ########## Functions #############
# ##################################


def tree_to_download_list(tree_array: list[Treeitem], rel_path: str = "") -> list:
    """Parse tree structure and return a list of files to download with relative
        paths to the base URL. It's meant to be used as recursive funciton to iter
        through the tree structure.

    Args:
        tree_array (list[TreeItem]): input array from tree JSON structure.
        rel_path (str, optional): relative path to resolve from. Defaults to "".

    Returns:
        list: list of files paths relative to the base URL.
    """
    li_files = []

    if not isinstance(tree_array, (list, tuple)):
        return li_files

    for item in tree_array:
        if item.get("type") == "directory":
            if item.get("name") != ".":
                new_rel_path = f"{rel_path}/{item.get('name')}"
            else:
                new_rel_path = f"{item.get('name')}"

        if "contents" in item:
            li_files.extend(
                tree_to_download_list(
                    tree_array=item.get("contents"),
                    rel_path=new_rel_path,
                )
            )
        elif item.get("type") == "file":
            li_files.append(f"{rel_path}/{item.get('name')}")
        else:
            logger.debug(f"Unsupported item type: {item.get('type')}")

    return li_files



#! python3  # noqa: E265

"""
    Download remote QGIS profiles to QDT working folder.

    Author: Julien Moura (https://github.com/guts)
"""


# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import logging

# package
from qgis_deployment_toolbelt.jobs.generic_job import GenericJob
from qgis_deployment_toolbelt.profiles import LocalGitHandler, RemoteGitHandler
from qgis_deployment_toolbelt.profiles.remote_http_handler import HttpHandler

# #############################################################################
# ########## Globals ###############
# ##################################

# logs
logger = logging.getLogger(__name__)


# #############################################################################
# ########## Classes ###############
# ##################################


class JobProfilesDownloader(GenericJob):
    """
    Job to download remote profiles and set them.
    """

    ID: str = "qprofiles-downloader"
    OPTIONS_SCHEMA: dict = {
        "branch": {
            "type": str,
            "required": False,
            "default": "master",
            "possible_values": None,
            "condition": None,
        },
        "protocol": {
            "type": str,
            "required": True,
            "default": "git_remote",
            "possible_values": ("git", "git_local", "git_remote", "http"),
            "condition": "in",
        },
        "source": {
            "type": str,
            "required": True,
            "default": None,
            "possible_values": ("https://", "http://", "git://", "file://"),
            "condition": "startswith",
        },
    }
    PROFILES_NAMES_DOWNLOADED: list = []

    def __init__(self, options: dict) -> None:
        """Instantiate the class.

        Args:
            options (List[dict]): list of dictionary with environment variables to set
            or remove.
        """
        super().__init__()
        self.options: dict = self.validate_options(options)

        # where QDT downloads remote repositories
        self.qdt_downloaded_repositories.mkdir(exist_ok=True, parents=True)
        logger.debug(f"Local repositories folder: {self.qdt_downloaded_repositories}")

    def run(self) -> None:
        """Execute job logic."""
        # prepare remote source
        if self.options.get("protocol") in ("git", "git_local", "git_remote"):
            if self.options.get("protocol") == "git":
                logger.warning(
                    DeprecationWarning(
                        "'git' protocol has been split into 2 more explicit: 'git_local' "
                        "(for git repositories accessible directly through file system "
                        "and 'git_remote' (for repositories accessible network request "
                        "to a remote git server through HTTP). "
                        "Please update your scenario consequently."
                    )
                )
            if self.options.get("protocol") == "git_remote" or self.options.get(
                "source"
            ).startswith(("git://", "http://", "https://")):
                downloader = RemoteGitHandler(
                    source_repository_url=self.options.get("source"),
                    branch_to_use=self.options.get("branch", "master"),
                )
            elif self.options.get("source").startswith("file://"):
                downloader = LocalGitHandler(
                    source_repository_path_or_uri=self.options.get("source"),
                    branch_to_use=self.options.get("branch", "master"),
                )
            else:
                logger.error(
                    f"Source type is not implemented yet: {self.options.get('source')}"
                    f"for '{self.options.get('protocol')}' protocol"
                )
                raise NotImplementedError
        elif self.options.get("protocol") == "http":
            if not self.options.get("source").startswith(("http://", "https://")):
                logger.error(
                    f"Source type not implemented yet: {self.options.get('source')} "
                    f"for '{self.options.get('protocol')}' protocol"
                )
                raise NotImplementedError
            downloader = HttpHandler(
                source_repository_path_or_uri=self.options.get("source"),
            )
        else:
            logger.critical(
                f"Protocol '{self.options.get('protocol')}' is not part of supported ones: "
                f"{self.OPTIONS_SCHEMA.get('protocol').get('possible_values')}"
            )
            raise NotImplementedError

        # run download operation
        downloader.download(destination_local_path=self.qdt_downloaded_repositories)

        # check of there are some profiles folders within the downloaded folder
        profiles_folders = self.list_downloaded_profiles()
        if profiles_folders is None:
            logger.error("No QGIS profile found in the downloaded folder.")
            return

        # store downloaded profiles names
        self.PROFILES_NAMES_DOWNLOADED = [d.name for d in profiles_folders]
        logger.info(
            f"{len(self.PROFILES_NAMES_DOWNLOADED)} downloaded profiles: "
            f"{', '.join(self.PROFILES_NAMES_DOWNLOADED)}"
        )

        logger.debug(f"Job {self.ID} ran successfully.")


# #############################################################################
# ##### Stand alone program ########
# ##################################

if __name__ == "__main__":
    """Standalone execution."""
    pass



#! python3  # noqa: E265

"""
    Tools to find installed QGIS version

    Author: Jean-Marie KERLOCH (https://github.com/jmkerloch)
"""

# #############################################################################
# ########## Libraries #############
# ##################################


# Standard library
import logging
import os
import subprocess
from os import environ, getenv
from os.path import expanduser, expandvars
from pathlib import Path
from shutil import which
from sys import platform as opersys

# package
from qgis_deployment_toolbelt.constants import (
    RE_QGIS_FINDER_DIR,
    RE_QGIS_FINDER_VERSION,
)
from qgis_deployment_toolbelt.exceptions import QgisInstallNotFound
from qgis_deployment_toolbelt.jobs.generic_job import GenericJob
from qgis_deployment_toolbelt.utils.check_path import check_path_exists

# #############################################################################
# ########## Globals ###############
# ##################################


# logs
logger = logging.getLogger(__name__)

# #############################################################################
# ########## Classes ###############
# ##################################


class JobQgisInstallationFinder(GenericJob):
    """
    Class to find installed QGIS version
    """

    ID: str = "qgis-installation-finder"
    OPTIONS_SCHEMA: dict = {
        "if_not_found": {
            "type": str,
            "required": False,
            "default": "warning",
            "possible_values": ("warning", "error"),
            "condition": "in",
        },
        "version_priority": {
            "type": (list, str),
            "required": False,
            "default": None,
            "possible_values": None,
            "condition": None,
        },
        "search_paths": {
            "type": (list, str),
            "required": False,
            "default": (
                expandvars("%PROGRAMFILES%"),
                expandvars(
                    expanduser(getenv("QDT_OSGEO4W_INSTALL_DIR", "C:\\OSGeo4W"))
                ),
            ),
            "possible_values": None,
            "condition": None,
        },
    }

    def __init__(self, options: dict) -> None:
        """Instantiate the class.

        Args:
            options (dict): job options
        """

        super().__init__()
        self.options: dict = self.validate_options(options)

    def run(self) -> None:
        """Define QDT_QGIS_EXE_PATH for shortcut creation"""
        if not self.run_needed():
            return

        if opersys not in ("linux", "win32"):
            logger.error(f"This job does not support your operating system: {opersys}")
            return

        installed_qgis_path: str | None = self.get_installed_qgis_path()

        if installed_qgis_path:
            logger.debug(f"{self.ID} : QDT_QGIS_EXE_PATH is now {installed_qgis_path}")
            os.environ["QDT_QGIS_EXE_PATH"] = installed_qgis_path
        else:
            if self.options.get("if_not_found", "warning") == "error":
                raise QgisInstallNotFound()
            else:
                logger.warning("No QGIS installation found")

        # log
        logger.debug(f"Job {self.ID} ran successfully.")

    # -- INTERNAL LOGIC ------------------------------------------------------
    def run_needed(self) -> bool:
        """Check if running the job is needed.

        Returns:
            bool: return True if job must be run, False otherwise
        """
        if qgis_bin_path := getenv("QDT_QGIS_EXE_PATH"):
            if check_path_exists(input_path=qgis_bin_path, raise_error=False):
                version_str = self._get_qgis_bin_version(qgis_bin_path)
                if version_str:
                    logger.info(
                        f"QDT_QGIS_EXE_PATH defined and path {qgis_bin_path} exists for "
                        f"QGIS {version_str}. {self.ID} job is skipped. "
                    )
                    return False
                else:
                    logger.warning(
                        f"QDT_QGIS_EXE_PATH defined and path {qgis_bin_path} exists but "
                        "the QGIS version can't be defined. Check environment variable."
                    )
        logger.debug(
            "'QDT_QGIS_EXE_PATH' is not defined. "
            "Searching for QGIS executable is necessary."
        )
        return True

    def get_installed_qgis_path(self) -> str | None:
        """Get list of installed QGIS executables.

        Returns:
            str | None : installed qgis path
        """
        if opersys == "linux":
            found_versions = self._get_linux_installed_qgis_path()
        elif opersys == "win32":
            # Check for installed version in the default install directory
            found_versions = self._get_windows_installed_qgis_path()

        if len(found_versions) == 0:
            return None

        logger.debug(f"Found installed QGIS: {found_versions}")
        latest_version = self._get_latest_version_from_list(
            versions=list(found_versions.keys())
        )

        # Define used version from version priority
        version_priority: list[str] = []
        if "version_priority" in self.options:
            version_priority = self.options["version_priority"]

        # Add preferred qgis version on top of the list
        if "QDT_PREFERRED_QGIS_VERSION" in environ:
            version_priority.insert(0, environ["QDT_PREFERRED_QGIS_VERSION"])

        for version in version_priority:
            if latest_matching_version := self._get_latest_matching_version_path(
                found_versions=found_versions, version=version
            ):
                return latest_matching_version

        latest_qgis = found_versions[latest_version]
        if len(version_priority) != 0:
            # No version found in version priority
            version_priority_str = ",".join(self.options["version_priority"])
            logger.info(
                f"QGIS version(s) [{version_priority_str}] not found. Using most recent found version {latest_version} : {latest_qgis}"
            )

        return latest_qgis

    @staticmethod
    def _get_latest_version_from_list(versions: list[str]) -> str | None:
        """Get latest version from a list, OSGEO4W are last.

        Args:
            versions (list[str]): list of found version

        Returns:
            str | None: latest version, None if no version provided
        """
        if len(versions):
            used_version = versions
            used_version.sort(reverse=True)
            return used_version[0]

        return None

    @staticmethod
    def _get_latest_matching_version_path(
        found_versions: dict[str, str], version: str
    ) -> str | None:
        """Get latest version path matching a wanted version.

        Args:
            found_versions (dict[str, str]): dict of found versions
            version (str): wanted version

        Returns:
            str | None: latest matching version path, None if not found
        """
        match_versions = []
        for found_version in found_versions.keys():
            if found_version.startswith(version):
                match_versions.append(found_version)
        # Use latest version
        if latest_version := JobQgisInstallationFinder._get_latest_version_from_list(
            match_versions
        ):
            return found_versions[latest_version]
        return None

    @staticmethod
    def _get_qgis_versions_in_dir(
        search_dir: str, search_patterns: list[str], found_version: dict[str, str]
    ) -> None:
        """Get QGIS binary path from an install directory.

        Args:
            install_dir (str): install directory
            search_patterns (list[str]): list of search pattern for qgis binary
            found_version (dict[str, str]): updated dict of qgis binary path and version
        """
        matchs = []

        logger.debug(
            f"Searching for QGIS binary in {search_dir} with pattern {search_patterns}"
        )

        for pattern in search_patterns:
            matchs += [file for file in Path(search_dir).rglob(pattern)]

        for match in matchs:
            if match.is_file():
                JobQgisInstallationFinder._search_qgis_version_and_add_to_dict(
                    qgis_bin=str(match), found_version=found_version
                )

    def _get_search_paths_with_environment_variable(self) -> list[str]:
        """Get search_paths option with environment variable update

        Returns:
            list[str]: search_paths
        """
        # search_paths
        search_paths = []
        if "search_paths" in self.options:
            for path in self.options["search_paths"]:
                search_paths.append(expandvars(expanduser(getenv(path, path))))
        return search_paths

    def _get_windows_installed_qgis_path(self) -> dict[str, str]:
        """Get dict of installed QGIS version in common install directory

        Returns:
            dict[str, str]: dict of QGIS version and QGIS bin path
        """

        # Get list of search path
        search_paths = self._get_search_paths_with_environment_variable()
        if len(search_paths) == 0:
            # Program files
            prog_file_dir = expandvars("%PROGRAMFILES%")

            for dir_name in os.listdir(prog_file_dir):
                # Check if the directory name matches the pattern
                match = RE_QGIS_FINDER_DIR.match(dir_name)
                if match:
                    search_paths.append(os.path.join(prog_file_dir, dir_name))

            # OSGEO4W
            search_paths.append(environ.get("QDT_OSGEO4W_INSTALL_DIR", "C:\\OSGeo4W"))

        return JobQgisInstallationFinder._get_qgis_found_version_dict_from_search_paths(
            search_paths=search_paths,
            search_patterns=["qgis-bin.exe", "qgis-ltr-bin.exe"],
        )

    @staticmethod
    def _get_qgis_found_version_dict_from_search_paths(
        search_paths: list[str], search_patterns: list[str]
    ) -> dict[str, str]:
        """Define qgis found version dict from a list of search path
        If identical version are found in multiple path, the first version found in search_path is used.

        Args:
            search_paths (list[str]): list of search paths
            search_patterns (list[str]): list of search pattern for qgis binary

        Returns:
            dict[str, str]: dict of qgis binary path for qgis version
        """
        # We search reversed to have the version defined in priority with the first value
        found_version = {}
        for search_path in reversed(search_paths):
            JobQgisInstallationFinder._get_qgis_versions_in_dir(
                search_path, search_patterns, found_version
            )

        return found_version

    @staticmethod
    def _search_qgis_version_and_add_to_dict(
        qgis_bin: str, found_version: dict[str, str]
    ) -> None:
        """Search qgis version from qgis binary and add to found_version dict if found

        Args:
            qgis_bin (str): qgis binary path
            found_version (dict[str, str]): updated dict of qgis binary path and version
        """
        version_str = JobQgisInstallationFinder._get_qgis_bin_version(qgis_bin=qgis_bin)
        if version_str:
            logger.debug(f"QGIS version {version_str} found : {qgis_bin}")
            found_version[version_str] = qgis_bin
        else:
            logger.warning(f"Can't define QGIS version for '{qgis_bin}' file.")

    def _get_linux_installed_qgis_path(self) -> dict[str, str]:
        """Get install qgis path for linux operating system with which

        Returns:
            dict[str, str]: dict of QGIS version and QGIS bin path
        """

        # search_paths
        search_paths = self._get_search_paths_with_environment_variable()

        found_version = (
            JobQgisInstallationFinder._get_qgis_found_version_dict_from_search_paths(
                search_paths=search_paths, search_patterns=["qgis"]
            )
        )

        # use which to find installed qgis
        if qgis_bin := which("qgis"):
            logger.debug(f"QGIS path found using which: {qgis_bin}")
            JobQgisInstallationFinder._search_qgis_version_and_add_to_dict(
                qgis_bin=qgis_bin, found_version=found_version
            )

        return found_version

    @staticmethod
    def _get_qgis_bin_version(qgis_bin: str) -> str | None:
        """Get QGIS bin version with --version

        Args:
            qgis_bin (str): QGIS bin path

        Returns:
            str | None: SemVer version, None if version not found
        """
        process = subprocess.Popen(
            f'"{qgis_bin}" --version', stdout=subprocess.PIPE, shell=True
        )
        stdout_, _ = process.communicate()
        version_str = stdout_.decode()
        version_match = RE_QGIS_FINDER_VERSION.match(version_str)
        if version_match:
            return version_match.group(1)
        return None



#! python3  # noqa E265

"""
    Usage from the repo root folder:

    .. code-block:: bash
        # for whole tests
        python -m unittest tests.test_utils_check_path
        # for specific test
        python -m unittest tests.test_utils_check_path.TestUtilsCheckPath.test_check_path_as_str_ok
"""


# standard library
import stat
import tempfile
import unittest
from os import chmod, getenv
from pathlib import Path

# project
from qgis_deployment_toolbelt.utils.check_path import (
    check_folder_is_empty,
    check_path,
    check_path_exists,
    check_path_is_readable,
    check_path_is_writable,
    check_var_can_be_path,
)

# ############################################################################
# ########## Classes #############
# ################################


class TestUtilsCheckPath(unittest.TestCase):
    """Test package metadata."""

    @classmethod
    def tearDownClass(cls) -> None:
        Path("tests/fixtures/tmp_file_no_readable.txt").unlink(missing_ok=True)
        Path("tests/fixtures/tmp_file_no_writable.txt").unlink(missing_ok=True)
        return super().tearDownClass()

    def test_check_path_as_str_ok(self):
        """Test filepath as str is converted into Path."""
        self.assertTrue(
            check_var_can_be_path(input_var="/this/is/an/imaginary/file.imaginary")
        )

        self.assertTrue(
            check_var_can_be_path(input_var="%PROGRAMFILES%/QGIS/3_22/bin/qgis-bin.exe")
        )

    def test_check_path_as_str_ko(self):
        """Test filepath from int can't be converted into Path."""
        with self.assertRaises(TypeError):
            check_var_can_be_path(input_var=1000)
        # no exception but False
        self.assertFalse(check_var_can_be_path(input_var=1000, raise_error=False))

    def test_check_path_exists_ok(self):
        """Test path exists."""
        # a valid Path instance pointing to an existing file
        self.assertTrue(check_path_exists(input_path=Path("setup.py")))
        # str is valid and point to an existing file
        self.assertTrue(check_path_exists(input_path="setup.py"))
        # str with user expand is valid and point to an existing folder
        self.assertTrue(check_path_exists(input_path="~"))

    def test_check_path_exists_ko(self):
        """Test path exists fail cases."""
        # str is not valid
        with self.assertRaises(TypeError):
            check_path_exists(input_path=1000)
        self.assertFalse(check_path_exists(input_path=1000, raise_error=False))
        # str is valid but not an existing file
        with self.assertRaises(FileExistsError):
            check_path_exists(input_path="/this/is/an/imaginary/file.imaginary")
        # no exception but False
        self.assertFalse(
            check_path_exists(
                input_path="/this/is/an/imaginary/file.imaginary", raise_error=False
            )
        )

    def test_check_path_readable_ok(self):
        """Test path is readable."""
        # a valid Path instance pointing to an existing file which is readable
        self.assertTrue(check_path_is_readable(input_path=Path("setup.py")))
        # str is valid and point to an existing file which is readable
        self.assertTrue(check_path_is_readable(input_path="setup.py"))

    def test_check_path_readable_ko(self):
        """Test path is readable fail cases."""
        # str is not valid
        with self.assertRaises(TypeError):
            check_path_is_readable(input_path=1000)
        self.assertFalse(check_path_is_readable(input_path=1000, raise_error=False))

    @unittest.skipIf(
        getenv("CI"), "Creating file on CI with specific rights is not working."
    )
    def test_check_path_readable_ko_specific(self):
        """Test path is readable fail cases."""
        # temporary fixture
        with tempfile.TemporaryDirectory(
            prefix="QDT_test_check_path",
            ignore_cleanup_errors=True,
        ) as tmpdirname:
            unreadable_file = Path(tmpdirname).joinpath("tmp_file_no_readable.txt")
            unreadable_file.touch(mode=0o333, exist_ok=True)

            # no exception but False
            self.assertFalse(
                check_path_is_readable(input_path=unreadable_file, raise_error=False)
            )

    def test_check_path_writable_ok(self):
        """Test path is writable."""
        # a valid Path instance pointing to an existing file which is writable
        self.assertTrue(check_path_is_writable(input_path=Path("setup.py")))
        # str is valid and point to an existing file which is writable
        self.assertTrue(check_path_is_writable(input_path="setup.py"))

    def test_check_path_writable_ko(self):
        """Test path is writable fail cases."""
        # str is not valid
        with self.assertRaises(TypeError):
            check_path_exists(input_path=1000)
        self.assertFalse(check_path_is_writable(input_path=1000, raise_error=False))

    def test_check_path_writable_ko_specific(self):
        """Test path is writable fail cases (specific)."""
        with tempfile.TemporaryDirectory(
            prefix="QDT_test_check_path",
            ignore_cleanup_errors=True,
        ) as tmpdirname:
            unwritable_file = Path(tmpdirname).joinpath("tmp_file_no_writable.txt")
            unwritable_file.touch(exist_ok=True)
            unwritable_file.write_text("this content is the last to be written here")

            chmod(unwritable_file, stat.S_IROTH)

            # str not valid, an existing file but not writable
            with self.assertRaises(IOError):
                check_path_is_writable(input_path=unwritable_file)

            # no exception but False
            self.assertFalse(
                check_path_is_writable(input_path=unwritable_file, raise_error=False)
            )

    def test_check_path_meta_ok(self):
        """Test meta check path."""
        # an existing file
        check_path(
            input_path="requirements.txt",
            must_be_a_file=True,
            must_be_a_folder=False,
        )
        check_path(
            input_path=Path("requirements.txt"),
            must_be_a_file=True,
            must_be_a_folder=False,
        )

        # an existing folder
        check_path(
            input_path=Path(__file__).parent,
            must_be_a_file=False,
            must_be_a_folder=True,
        )

    def test_check_path_meta_ko(self):
        """Test meta check path fail cases."""
        # invalid path
        with self.assertRaises(TypeError):
            check_path(
                input_path=1000,
            )
        self.assertFalse(check_path(input_path=1000, raise_error=False))

        # mutual exclusive options
        with self.assertRaises(ValueError):
            check_path(
                input_path="requirements.txt",
                must_be_a_file=True,
                must_be_a_folder=True,
            )

        # must exist
        self.assertFalse(
            check_path(
                input_path="imaginary/path", raise_error=False, must_exists=False
            )
        )
        self.assertFalse(
            check_path(input_path="imaginary/path", raise_error=False, must_exists=True)
        )
        with self.assertRaises(FileExistsError):
            check_path(input_path="imaginary/path", must_exists=True)

        # must be readable
        self.assertTrue(
            check_path(
                input_path="imaginary/path",
                raise_error=False,
                must_exists=False,
                must_be_readable=False,
            )
        )
        self.assertTrue(
            check_path(
                input_path="setup.py",
                raise_error=False,
                must_exists=False,
                must_be_readable=True,
            )
        )

        # must be writable
        self.assertFalse(
            check_path(
                input_path="imaginary/path",
                raise_error=False,
                must_exists=False,
                must_be_readable=False,
                must_be_writable=True,
            )
        )
        self.assertTrue(
            check_path(
                input_path=f"tests/{Path(__file__).name}",
                raise_error=True,
                must_exists=False,
                must_be_readable=False,
                must_be_writable=True,
            )
        )

        # must be a file
        self.assertFalse(
            check_path(
                input_path=Path(__file__).parent,
                raise_error=False,
                must_exists=True,
                must_be_a_file=True,
            )
        )
        with self.assertRaises(FileNotFoundError):
            check_path(
                input_path=Path(__file__).parent,
                raise_error=True,
                must_exists=True,
                must_be_a_file=True,
            )

        # must be a folder
        self.assertFalse(
            check_path(
                input_path=Path(__file__),
                raise_error=False,
                must_exists=True,
                must_be_a_folder=True,
            )
        )
        with self.assertRaises(NotADirectoryError):
            check_path(
                input_path=Path(__file__),
                raise_error=True,
                must_exists=True,
                must_be_a_folder=True,
            )

    @unittest.skipIf(
        getenv("CI"), "Creating file on CI with specific rights is not working."
    )
    def test_check_path_meta_ko_specific(self):
        """Test meta check path is readbale / writable fail cases (specific)."""
        # temporary fixture
        not_writable_file = Path("tests/fixtures/tmp_file_no_writable.txt")
        not_writable_file.touch()
        chmod(not_writable_file, stat.S_IREAD | stat.S_IROTH)

        # str not valid, an existing file but not writable
        with self.assertRaises(IOError):
            check_path(
                input_path=not_writable_file, must_be_a_file=True, must_be_writable=True
            )

        # no exception but False
        self.assertFalse(
            check_path(
                input_path=not_writable_file,
                must_be_a_file=True,
                must_be_writable=True,
                raise_error=False,
            )
        )

        not_writable_file.unlink()

    def test_check_folder_is_empty(self):
        """Test empty folder recognition."""
        with tempfile.TemporaryDirectory(
            prefix="QDT_test_check_path_",
            ignore_cleanup_errors=True,
            suffix="_empty_folder",
        ) as tmpdirname:
            self.assertTrue(check_folder_is_empty(Path(tmpdirname)))

        with tempfile.TemporaryDirectory(
            prefix="QDT_test_check_path_",
            ignore_cleanup_errors=True,
            suffix="_not_empty_folder",
        ) as tmpdirname:
            Path(tmpdirname).joinpath(".gitkeep").touch()
            self.assertFalse(check_folder_is_empty(Path(tmpdirname)))

        with self.assertRaises(TypeError):
            check_folder_is_empty(input_var=1000)
        # no exception but False
        self.assertFalse(check_folder_is_empty(input_var=1000, raise_error=False))


# ############################################################################
# ####### Stand-alone run ########
# ################################
if __name__ == "__main__":
    unittest.main()



#! python3  # noqa: E265

"""
    Custom INI files interpolation.

    Author: Julien Moura (https://github.com/guts)
"""


# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import logging
from collections import ChainMap
from configparser import BasicInterpolation, ConfigParser, InterpolationSyntaxError
from os.path import expanduser, expandvars

# #############################################################################
# ########## Globals ###############
# ##################################

# logs
logger = logging.getLogger(__name__)

# #############################################################################
# ########## Classes ###############
# ##################################


class EnvironmentVariablesInterpolation(BasicInterpolation):
    """Custom interpolation to handle environment variables in ini files.

    Inspired from https://gist.github.com/malexer/ee2f93b1973120925e8beb3f36b184b8.


    Example:

        .. code-block:: ini

            [test]
            user = $USER
            user_home = $HOME
            fake_value_from_environment_variable = $QDT_TEST_ENV_VARIABLE
            pictures = %(user_home)s/.cache/qgis-deployment-toolbelt
    """

    def before_get(
        self,
        parser: ConfigParser,
        section: str,
        option: str,
        value: str,
        defaults: ChainMap,
    ) -> str:
        """Called for every get option=value line in INI file.

        Args:
            parser (ConfigParser): parser whose function is overloaded
            section (str): section's name
            option (str): option's name
            value (str): value to try to interpolate
            defaults (ChainMap): defaults options/values

        Returns:
            str: interpolated value
        """
        # Add try catch because QGIS INI file can omit some % escaping
        try:
            value = super().before_get(parser, section, option, value, defaults)
        except InterpolationSyntaxError:
            return value

        try:
            return expandvars(expanduser(value))
        except Exception as exc:
            logger.error(
                f"Failed to interpolate {value} in {section}/{option}. Trace: {exc}"
            )
            return value

    def before_set(
        self,
        parser: ConfigParser,
        section: str,
        option: str,
        value: str,
    ) -> str:
        """Called for every set option=value line in INI file.

        Args:
            parser (ConfigParser): parser whose function is overloaded
            section (str): section's name
            option (str): option's name
            value (str): value to try to interpolate
        Returns:
            str: interpolated value
        """
        # Add try catch because QGIS INI file can omit some % escaping
        try:
            return super().before_set(parser, section, option, value)
        except (InterpolationSyntaxError, ValueError):
            return value

    def before_write(
        self,
        parser: ConfigParser,
        section: str,
        option: str,
        value: str,
    ) -> str:
        """Called before write option=value line in INI file.

        Args:
            parser (ConfigParser): parser whose function is overloaded
            section (str): section's name
            option (str): option's name
            value (str): value to try to interpolate

        Returns:
            str: interpolated value
        """
        value = super().before_write(parser, section, option, value)
        try:
            return expandvars(expanduser(value))
        except Exception as exc:
            logger.error(
                f"Failed to interpolate {value} in {section}/{option}. Trace: {exc}"
            )
            return value


# #############################################################################
# ##### Stand alone program ########
# ##################################

if __name__ == "__main__":
    """Standalone execution."""
    pass



#! python3  # noqa: E265

"""
    Metadata bout the package to easily retrieve informations about it.
    See: https://packaging.python.org/guides/single-sourcing-package-version/
"""

from datetime import date

__all__ = [
    "__author__",
    "__copyright__",
    "__email__",
    "__license__",
    "__summary__",
    "__title__",
    "__uri__",
    "__version__",
]

__author__ = (
    "Julien Moura (Oslandia), Vincent Bré (Oslandia), Jean-Marie Kerloch (Oslandia)"
)
__copyright__ = f"2021 - {date.today().year}, {__author__}"
__email__ = "qgis@oslandia.com"
__executable_name__ = "qgis-deployment-toolbelt"
__package_name__ = "qgis-deployment-toolbelt"
__keywords__ = ["cli", "deployment", "profiles", "qdt", "QGIS"]
__license__ = "Apache-2.0"
__summary__ = (
    "QGIS deployment toolbelt is a CLI (Command Line Interface) to perform "
    "redundant operations after a QGIS deployment."
)
__title__ = "QGIS Deployment Toolbelt"
__title_clean__ = "".join(e for e in __title__ if e.isalnum())
__uri_homepage__ = "https://qgis-deployment.github.io/qgis-deployment-toolbelt-cli/"
__uri_repository__ = "https://github.com/qgis-deployment/qgis-deployment-toolbelt-cli/"
__uri_tracker__ = f"{__uri_repository__}issues/"
__uri__ = __uri_repository__

__version__ = "0.36.2"
__version_info__ = tuple(
    [
        int(num) if num.isdigit() else num
        for num in __version__.replace("-", ".", 1).split(".")
    ]
)



#! python3  # noqa E265

"""
    Usage from the repo root folder:

    .. code-block:: bash
        # for whole tests
        python -m unittest tests.test_qplugin_object
        # for specific test
        python -m unittest tests.test_qplugin_object.TestQgisPluginObject.test_profile_load_from_json_basic
"""

# standard
import unittest
from pathlib import Path
from shutil import rmtree

# project
from qgis_deployment_toolbelt.plugins.plugin import QgisPlugin
from qgis_deployment_toolbelt.profiles.qdt_profile import QdtProfile
from qgis_deployment_toolbelt.utils.file_downloader import download_remote_file_to_local

# ############################################################################
# ########## Classes #############
# ################################


class TestQgisPluginObject(unittest.TestCase):
    """Test QGIS Plugin abstraction class."""

    # -- Standard methods --------------------------------------------------------
    @classmethod
    def setUpClass(cls):
        """Executed when module is loaded before any test."""
        cls.good_profiles_files = sorted(
            Path("tests/fixtures/").glob("profiles/good_*.json")
        )

        # -- Download a plugin to perfom tests
        # prepare local download path
        cls.sample_plugin_downloaded = Path(
            "tests/fixtures/tmp/qtribu/2733_qtribu_0-14-2.zip"
        )
        cls.sample_plugin_downloaded.parent.mkdir(parents=True, exist_ok=True)

        # download plugin zip archive (only if it doesn't exist already)
        if not cls.sample_plugin_downloaded.exists():
            download_remote_file_to_local(
                remote_url_to_download="https://github.com/geotribu/qtribu/releases/download/0.14.2/qtribu.0.14.2.zip",
                local_file_path=cls.sample_plugin_downloaded.resolve(),
                content_type="application/zip",
            )

        # check download worked
        assert (
            cls.sample_plugin_downloaded.is_file() is True
        ), "Downloading fixture plugin failed."

    @classmethod
    def tearDownClass(cls) -> None:
        rmtree(path=cls.sample_plugin_downloaded.parent, ignore_errors=True)

    def test_qplugin_load_from_profile(self):
        """Test plugin object loading from profile object."""
        for p in self.good_profiles_files:
            qdt_profile = QdtProfile.from_json(
                profile_json_path=p,
                profile_folder=p.parent,
            )

            self.assertIsInstance(qdt_profile.plugins, list)
            if len(qdt_profile.plugins):
                for i in qdt_profile.plugins:
                    self.assertIsInstance(i, QgisPlugin)

    def test_qplugin_load_from_dict(self):
        """Test plugin object loading from JSON."""
        sample_plugin_complete = {
            "name": "french_locator_filter",
            "version": "1.0.4",
            "url": "https://plugins.qgis.org/plugins/french_locator_filter/version/1.0.4/download/",
            "location": "remote",
        }

        plugin_obj_one: QgisPlugin = QgisPlugin.from_dict(sample_plugin_complete)

        sample_plugin_incomplete = {
            "name": "french_locator_filter",
            "version": "1.0.4",
            "official_repository": True,
        }

        plugin_obj_two: QgisPlugin = QgisPlugin.from_dict(sample_plugin_incomplete)

        self.assertEqual(plugin_obj_one, plugin_obj_two)
        self.assertEqual(plugin_obj_one.download_url, plugin_obj_one.uri_to_zip)

    def test_qplugin_load_from_dict_local(self):
        """Test plugin object loading from dict, pointing to a local plugin."""
        # plugin as dict
        plugin_dict_local = {
            "name": "QTribu",
            "folder_name": "qtribu",
            "official_repository": False,
            "plugin_id": 2733,
            "location": "local",
            "url": self.sample_plugin_downloaded,
            "version": "0.14.2",
        }

        # plugin as object
        plugin_obj: QgisPlugin = QgisPlugin.from_dict(plugin_dict_local)
        self.assertIsInstance(plugin_obj.uri_to_zip, Path)

        # using a file:// prefix
        plugin_dict_local["url"] = f"file://{self.sample_plugin_downloaded}"
        plugin_obj: QgisPlugin = QgisPlugin.from_dict(plugin_dict_local)

    def test_qplugin_load_from_zip(self):
        """Test plugin object loading from a ZIP archive downloaded."""
        # plugin as dict
        sample_plugin_complex = {
            "name": "Layers menu from project",
            "version": "2.1.0",
            "url": "https://plugins.qgis.org/plugins/menu_from_project/version/2.1.0/download/",
            "location": "remote",
            "plugin_id": 1846,
        }
        # plugin as object
        plugin_obj: QgisPlugin = QgisPlugin.from_dict(sample_plugin_complex)

        # some attributes are not set
        self.assertIsNone(plugin_obj.folder_name)

        # prepare local download path
        local_plugin_download = Path(
            f"{self.sample_plugin_downloaded.parent.parent}/{plugin_obj.installation_folder_name}/"
            f"{plugin_obj.id_with_version}.zip"
        )

        # some attributes have been set
        self.assertIsNotNone(plugin_obj.folder_name)

        # download plugin zip archive
        download_remote_file_to_local(
            remote_url_to_download=plugin_obj.download_url,
            local_file_path=local_plugin_download.resolve(),
            content_type="application/zip",
        )

        # check download worked
        self.assertTrue(local_plugin_download.is_file())

        # load from zip
        plugin_obj_from_zip: QgisPlugin = QgisPlugin.from_zip(
            input_zip_path=local_plugin_download
        )

        # check types
        self.assertIsInstance(plugin_obj_from_zip.name, str)
        self.assertIsInstance(plugin_obj_from_zip.version, str)
        self.assertIsInstance(plugin_obj_from_zip.folder_name, str)
        self.assertIsInstance(plugin_obj_from_zip.qgis_maximum_version, str)
        self.assertIsInstance(plugin_obj_from_zip.qgis_minimum_version, str)

        # check values
        self.assertEqual(plugin_obj.name, plugin_obj_from_zip.name)
        self.assertEqual(plugin_obj.version, plugin_obj_from_zip.version)
        self.assertEqual(plugin_obj.folder_name, plugin_obj_from_zip.folder_name)
        self.assertEqual(plugin_obj.url, sample_plugin_complex.get("url"))
        self.assertEqual(plugin_obj.download_url, sample_plugin_complex.get("url"))
        self.assertEqual(plugin_obj.uri_to_zip, sample_plugin_complex.get("url"))

        # clean up
        rmtree(path=local_plugin_download.parent, ignore_errors=True)

    def test_qplugin_versions_comparison_semver(self):
        """Test plugin compare versions semver"""
        plugin_v1: QgisPlugin = QgisPlugin.from_dict(
            {
                "name": "Sample plugin",
                "version": "1.0.0",
            }
        )

        plugin_v2: QgisPlugin = QgisPlugin.from_dict(
            {
                "name": "Sample plugin",
                "version": "1.1.0",
            }
        )

        plugin_v3: QgisPlugin = QgisPlugin.from_dict(
            {
                "name": "Sample plugin",
                "version": "3.1.0",
            }
        )

        self.assertTrue(plugin_v1.is_older_than(plugin_v2.version))
        self.assertTrue(plugin_v1.is_older_than(plugin_v2))
        self.assertTrue(plugin_v1.is_older_than(plugin_v3.version))
        self.assertTrue(plugin_v1.is_older_than(plugin_v3))
        self.assertTrue(plugin_v2.is_older_than(plugin_v3.version))
        self.assertTrue(plugin_v2.is_older_than(plugin_v3))
        self.assertFalse(plugin_v2.is_older_than(plugin_v1))

    def test_qplugin_versions_comparison_semver_prefixed(self):
        """Test plugin compare versions semver with a prefix"""
        plugin_v1: QgisPlugin = QgisPlugin.from_dict(
            {
                "name": "Sample plugin",
                "version": "v1.0.0",
            }
        )

        plugin_v2: QgisPlugin = QgisPlugin.from_dict(
            {
                "name": "Sample plugin",
                "version": "v1.1.0",
            }
        )

        plugin_v3: QgisPlugin = QgisPlugin.from_dict(
            {
                "name": "Sample plugin",
                "version": "v3.1.0",
            }
        )

        self.assertTrue(plugin_v1.is_older_than(plugin_v2.version))
        self.assertTrue(plugin_v1.is_older_than(plugin_v2))
        self.assertTrue(plugin_v1.is_older_than(plugin_v3.version))
        self.assertTrue(plugin_v1.is_older_than(plugin_v3))
        self.assertTrue(plugin_v2.is_older_than(plugin_v3.version))
        self.assertTrue(plugin_v2.is_older_than(plugin_v3))
        self.assertFalse(plugin_v2.is_older_than(plugin_v1))

    def test_qplugin_versions_comparison_calver(self):
        """Test plugin compare versions calver"""
        plugin_v1: QgisPlugin = QgisPlugin.from_dict(
            {"name": "Sample plugin", "version": "2021.9.10"}
        )

        plugin_v2: QgisPlugin = QgisPlugin.from_dict(
            {
                "name": "Sample plugin",
                "version": "2021.12.10",
            }
        )

        plugin_v3: QgisPlugin = QgisPlugin.from_dict(
            {
                "name": "Sample plugin",
                "version": "2023.2.10",
            }
        )

        self.assertTrue(plugin_v1.is_older_than(plugin_v2.version))
        self.assertTrue(plugin_v1.is_older_than(plugin_v2))
        self.assertTrue(plugin_v1.is_older_than(plugin_v3.version))
        self.assertTrue(plugin_v1.is_older_than(plugin_v3))
        self.assertTrue(plugin_v2.is_older_than(plugin_v3.version))
        self.assertTrue(plugin_v2.is_older_than(plugin_v3))
        self.assertFalse(plugin_v2.is_older_than(plugin_v1))

    def test_qplugin_versions_comparison_bad(self):
        """Test plugin compare versions issues"""
        plugin_v1: QgisPlugin = QgisPlugin.from_dict(
            {"name": "Sample plugin", "version": "2021.9.10"}
        )

        plugin_v2: QgisPlugin = QgisPlugin.from_dict(
            {
                "name": "Sample plugin 2",
                "version": "2021.12.10",
            }
        )

        self.assertTrue(plugin_v1.is_older_than(plugin_v2.version))
        self.assertFalse(plugin_v2.is_older_than(plugin_v1))

        plugin_bad_version: QgisPlugin = QgisPlugin.from_dict(
            {
                "name": "Sample plugin",
                "version": "march 2022",
            }
        )

        self.assertIsNone(plugin_v1.is_older_than(plugin_bad_version))


# ############################################################################
# ####### Stand-alone run ########
# ################################
if __name__ == "__main__":
    unittest.main()



#! python3  # noqa: E265

"""
    Read and write QGIS configuration files.

    Author: Julien Moura (https://github.com/guts)
"""


# #############################################################################
# ########## Libraries #############
# ##################################

# special
from __future__ import annotations

# Standard library
import logging
from pathlib import Path
from typing import Literal

# package
from qgis_deployment_toolbelt.utils.check_path import check_path
from qgis_deployment_toolbelt.utils.ini_interpolation import (
    EnvironmentVariablesInterpolation,
)
from qgis_deployment_toolbelt.utils.ini_parser_with_path import CustomConfigParser
from qgis_deployment_toolbelt.utils.win32utils import normalize_path

# #############################################################################
# ########## Globals ###############
# ##################################

# logs
logger = logging.getLogger(__name__)

# #############################################################################
# ########## Classes ###############
# ##################################


class QgisIniHelper:
    """Helper to manipulate QGIS configuration files (*.ini)."""

    SUPPORTED_INI_TYPES: tuple[str, str, str] = (
        "plugin_metadata",
        "profile_qgis3",
        "profile_qgis3customization",
    )

    ini_type: str | None = None

    def __init__(
        self,
        ini_filepath: Path,
        ini_type: Literal[
            "profile_qgis3", "profile_qgis3customization", "plugin_metadata", None
        ] = None,
        strict: bool = False,
        enable_environment_variables_interpolation: bool = True,
    ) -> None:
        """Instanciation.

        Args:
            ini_filepath (Path): path to the QGIS3.ini configuration file
            ini_type (Literal[ &quot;profile_qgis3&quot;, \
                &quot;profile_qgis3customization&quot;, &quot;plugin_metadata&quot;, \
                None ], optional): type of ini file. None enables autodetection. \
                Defaults to None.
            strict (bool, optional): strict mode applied to ConfigParser. Defaults to \
                False.
            enable_environment_variables_interpolation (bool, optional): if enabled, \
                values matching environment variables are interepreted. Defaults to True.
        """
        if (
            ini_filepath is not None
            and ini_type is not None
            and ini_type in self.SUPPORTED_INI_TYPES
        ):
            self.ini_type = ini_type
            if self.ini_type == "profile_qgis3":
                self.profile_config_path = ini_filepath
                self.profile_customization_path = ini_filepath.with_name(
                    "QGISCUSTOMIZATION3.ini"
                )
            elif self.ini_type == "profile_qgis3customization":
                self.profile_config_path = ini_filepath.with_name("QGIS3.ini")
                self.profile_customization_path = ini_filepath
            else:
                self.profile_config_path = None
                self.profile_customization_path = None
        elif ini_filepath.stem == "QGIS3":
            self.ini_type = "profile_qgis3"
            self.profile_config_path = ini_filepath
            self.profile_customization_path = ini_filepath.with_name(
                "QGISCUSTOMIZATION3.ini"
            )
        elif ini_filepath.stem == "QGISCUSTOMIZATION3":
            self.ini_type = "profile_qgis3customization"
            self.profile_config_path = ini_filepath.with_name("QGIS3.ini")
            self.profile_customization_path = ini_filepath
        elif ini_filepath.name == "metadata.txt":
            self.ini_type = "plugin_metadata"
            self.profile_config_path = None
            self.profile_customization_path = None
        else:
            logger.warning(f"Unrecognized ini type: {ini_filepath}")

        # check if file exists
        if not check_path(
            input_path=ini_filepath,
            must_be_a_file=True,
            must_be_readable=True,
            must_exists=True,
            raise_error=False,
        ):
            logger.info(f"The specified file does not exist: {ini_filepath.resolve()}.")
        self.ini_filepath = ini_filepath

        # store options
        self.strict_mode = strict
        self.enable_environment_variables_interpolation = (
            enable_environment_variables_interpolation
        )

    def cfg_parser(self) -> CustomConfigParser:
        """Return config parser with options for QGIS ini files.

        Returns:
            CustomConfigParser: config parser
        """
        qgis_cfg_parser = CustomConfigParser(
            strict=self.strict_mode,
            interpolation=(
                EnvironmentVariablesInterpolation()
                if self.enable_environment_variables_interpolation
                else None
            ),
        )
        qgis_cfg_parser.optionxform = str
        return qgis_cfg_parser

    # UI customization
    def is_ui_customization_enabled(
        self, ini_file: CustomConfigParser | Path | None = None
    ) -> bool | None:
        """Determine if UI customization is enabled.

        Args:
            ini_file (Union[CustomConfigParser, Path]): input ini file to check.
                A warning is raised if the filename is not QGIS3.ini.
                If None, self.profile_config_path is used.

        Returns:
            bool: True if customization is enabled
        """
        if ini_file is None and isinstance(self.profile_config_path, Path):
            logger.debug(
                "Using configuration file defined at object level: "
                f"{self.profile_config_path.resolve()}"
            )
            return self.is_ui_customization_enabled(ini_file=self.profile_config_path)

        if self.ini_type not in ("profile_qgis3", "profile_qgis3customization"):
            logger.debug(
                f"Invalid ini type: {self.ini_type}. Must a QGIS3.ini or a "
                "QGIS3CUSTOMIZATION.ini"
            )
            return None

        if isinstance(ini_file, CustomConfigParser):
            if ini_file.has_option(section="UI", option="Customization\\enabled"):
                return ini_file.getboolean(
                    section="UI", option="Customization\\enabled"
                )
            else:
                return False
        elif isinstance(ini_file, Path):
            if ini_file.stem == "QGIS3":
                logger.warning(
                    "Input file does not seem to be a QGIS profile configuration file "
                    f"(QGIS/QGIS3.ini): {ini_file}"
                )

            cfg_parser = self.cfg_parser()
            cfg_parser.read(ini_file, encoding="UTF8")
            return self.is_ui_customization_enabled(cfg_parser)
        elif (
            ini_file is None
            and isinstance(self.profile_config_path, Path)
            and self.profile_config_path.exists()
        ):
            cfg_parser = self.cfg_parser()
            cfg_parser.read([self.profile_config_path], encoding="UTF8")
            return self.is_ui_customization_enabled(cfg_parser)
        else:
            return False

    def is_splash_screen_set(
        self, ini_file: CustomConfigParser | Path | None = None
    ) -> bool | None:
        """Determine if a custom splash screen is set or not.

        Args:
            ini_file (Union[CustomConfigParser, Path]): input ini file to check.
                A warning is raised if the filename is not QGISCUSTOMIZATION3.ini.
                If None, self.profile_customization_path is used.

        Returns:
            bool: True if a splash is set
        """
        # if ini_file is None but defined at object level, let's use it
        if ini_file is None and isinstance(self.profile_customization_path, Path):
            logger.debug(
                "Using customization file defined at object level: "
                f"{self.profile_customization_path.resolve()}"
            )
            return self.is_splash_screen_set(ini_file=self.profile_customization_path)

        # if self.ini_type not in ("profile_qgis3", "profile_qgis3customization"):
        #     logger.debug(
        #         f"Invalid ini type: {self.ini_type}. Must a QGIS3.ini or a "
        #         "QGIS3CUSTOMIZATION.ini"
        #     )
        #     return None

        if isinstance(ini_file, CustomConfigParser):
            if ini_file.has_option(section="Customization", option="splashpath"):
                splash_path = ini_file.get(section="Customization", option="splashpath")
                if isinstance(splash_path, str) and not len(splash_path.strip()):
                    logger.info(
                        f"{ini_file.get_initial_file_path().resolve()} has a splashpath "
                        "option BUT the value set is NOT a valid path"
                    )
                    return False
                else:
                    logger.debug(
                        f"{ini_file.initial_file_path} has a splash screen set: {splash_path}"
                    )
                    return True
            else:
                logger.debug(
                    f"{ini_file.get_initial_file_path().resolve()} has no "
                    "splash path set."
                )
                return False
        elif isinstance(ini_file, Path) and check_path(
            input_path=ini_file,
            must_be_a_file=True,
            must_be_readable=True,
            must_exists=True,
            raise_error=False,
        ):
            cfg_parser = self.cfg_parser()
            cfg_parser.read(ini_file, encoding="UTF8")
            logger.debug(
                f"{ini_file} is an existing file and has been parsed. Let's check if a "
                "splash path is set."
            )
            return self.is_splash_screen_set(cfg_parser)
        else:
            return False

    def set_ui_customization_enabled(self, switch: bool = True) -> bool:
        """Enable/disable UI customization in the profile QGIS3.ini file.

        Args:
            switch (bool, optional): True to enable, False to disable UI customization.
                Defaults to True.

        Returns:
            bool: UI customization state. True is enabled, False is disabled.
        """
        # local variables
        section = "UI"
        option = "Customization\\enabled"

        # boolean syntax for PyQt
        switch_value = "false"
        if switch:
            switch_value = "true"

        # make sure that the file exists
        if not self.profile_config_path.exists():
            logger.warning(
                f"Configuration file {self.profile_config_path} doesn't exist. "
                "It will be created but maybe it was not the expected behavior."
            )
            self.profile_config_path.touch(exist_ok=True)
            self.profile_config_path.write_text(
                data=f"[UI]\nCustomization\\enabled={switch_value}", encoding="UTF8"
            )
            return switch

        # read configuration file
        ini_qgis3 = self.cfg_parser()
        ini_qgis3.read(self.profile_config_path, encoding="UTF8")

        # if section and option already exist
        if ini_qgis3.has_option(section=section, option=option):  # = section AND option
            actual_state = ini_qgis3.getboolean(section=section, option=option)
            # when actual UI customization is enabled
            if actual_state and switch:
                logger.debug(
                    f"UI Customization is already ENABLED in {self.profile_config_path}"
                )
                return True
            elif actual_state and not switch:
                ini_qgis3.set(
                    section=section,
                    option=option,
                    value=switch_value,
                )
                with self.profile_config_path.open("w", encoding="UTF8") as configfile:
                    ini_qgis3.write(configfile, space_around_delimiters=False)
                logger.debug(
                    "UI Customization was ENABLED and has been "
                    f"DISABLED in {self.profile_config_path}"
                )
                return False

            # when actual UI customization is disabled
            elif not actual_state and switch:
                ini_qgis3.set(
                    section=section,
                    option=option,
                    value=switch_value,
                )
                with self.profile_config_path.open("w", encoding="UTF8") as configfile:
                    ini_qgis3.write(configfile, space_around_delimiters=False)
                logger.debug(
                    "UI Customization was DISABLED and has been "
                    f"ENABLED in {self.profile_config_path}"
                )
                return True
            elif not actual_state and not switch:
                logger.debug(
                    f"UI Customization is already DISABLED in {self.profile_config_path}"
                )
                return True
        elif ini_qgis3.has_section(section=section) and switch:
            # section exist but not the option, so let's add it as required
            ini_qgis3.set(
                section=section,
                option=option,
                value=switch_value,
            )
            with self.profile_config_path.open("w", encoding="UTF8") as configfile:
                ini_qgis3.write(configfile, space_around_delimiters=False)
            logger.debug(
                f"'{option}' option was missing in {section}, it's just been added and "
                f"ENABLED in {self.profile_config_path}"
            )
            return True
        elif not ini_qgis3.has_section(section=section) and switch:
            # even the section is missing. Let's add everything
            ini_qgis3.add_section(section)
            ini_qgis3.set(
                section=section,
                option=option,
                value=switch_value,
            )
            with self.profile_config_path.open("w", encoding="UTF8") as configfile:
                ini_qgis3.write(configfile, space_around_delimiters=False)
            logger.debug(
                f"'{option}' option was missing in {section}, it's just been added and "
                f"ENABLED in {self.profile_config_path}"
            )
            return True
        else:
            logger.debug(
                f"Section '{section}' is not present so {option} is DISABLED in "
                f"{self.profile_config_path}"
            )
            return False

    def set_splash_screen(
        self,
        ini_file: CustomConfigParser | Path | None = None,
        splash_screen_filepath: Path | None = None,
        switch: bool = True,
    ) -> bool:
        """Add/remove splash screen path to the QGISCUSTOMIZATION3.ini file.

        Args:
            ini_file (Union[CustomConfigParser, Path]): input ini file to check.
                A warning is raised if the filename is not QGISCUSTOMIZATION3.ini.
                If None, self.profile_customization_path is used.
            splash_screen_filepath (Path, optional): path to the folder containing the
                splash.png file, defaults to None. Defaults to None.
            switch (bool, optional): True to add, False to remove, defaults to True.
                Defaults to True.

        Returns:
            bool: True is the splash folder is present. False is absent.
        """
        # if ini_file is None but defined at object level, let's use it
        if ini_file is None and isinstance(self.profile_customization_path, Path):
            logger.debug(
                "Using customization file defined at object level: "
                f"{self.profile_customization_path.resolve()}"
            )
            return self.set_splash_screen(
                ini_file=self.profile_customization_path, switch=switch
            )
        elif ini_file is None and self.profile_customization_path is None:
            raise ValueError(
                "Both passed ini file and the object's defined are not "
                "defined. Can't process."
            )

        # SWITCH=False and NOT DEFINED --> EXIT
        if not switch and not self.is_splash_screen_set(ini_file=ini_file):
            logger.debug(
                f"As required ({switch=}, splash screen is not set in {ini_file}"
            )
            return False

        # SWITCH=TRUE --> ENABLE
        if switch and isinstance(splash_screen_filepath, Path):
            # normalize splash screen folder path
            nrm_splash_screen_folder = normalize_path(
                splash_screen_filepath.parent, add_trailing_slash_if_dir=True
            )
            logger.debug(
                f"Splash screen path has been normalized: {nrm_splash_screen_folder}"
            )
        elif switch and splash_screen_filepath is None:
            logger.warning(f"{switch=} but splash screen filepath not defined")
        else:
            pass

        # make sure that the file exists
        if isinstance(ini_file, Path) and ini_file.exists():
            # open
            cfg_parser = self.cfg_parser()
            cfg_parser.read(ini_file, encoding="UTF8")
            logger.debug(
                f"{ini_file} is an existing file, has been parsed. Let's check if a "
                "splash path is set."
            )
            return self.set_splash_screen(
                ini_file=cfg_parser,
                splash_screen_filepath=splash_screen_filepath,
                switch=switch,
            )
        elif isinstance(ini_file, Path) and not ini_file.exists() and switch:
            logger.warning(
                f"Configuration file {ini_file} doesn't exist. "
                "It will be created but maybe it was not the expected behavior."
            )
            ini_file.parent.mkdir(parents=True, exist_ok=True)
            ini_file.touch(exist_ok=True)
            ini_file.write_text(
                data=f"[Customization]\nsplashpath={nrm_splash_screen_folder}",
                encoding="UTF8",
            )
            return switch
        elif isinstance(ini_file, Path) and not ini_file.exists() and not switch:
            logger.debug(f"'{ini_file} doesn't exist. So, no need to do anything.")
            return switch
        else:
            pass

        # FROM NOW: isinstance(ini_file, CustomConfigParser) is True
        qgiscustomization3ini_filepath = ini_file.get_initial_file_path()
        option = "splashpath"
        section = "Customization"

        # check existing option value
        if ini_file.has_option(section=section, option=option):
            actual_state = ini_file.get(section=section, option=option)

            if not switch:
                ini_file.remove_option(section=section, option=option)
                with qgiscustomization3ini_filepath.open(
                    "w", encoding="UTF8"
                ) as configfile:
                    ini_file.write(configfile, space_around_delimiters=False)
                logger.debug(
                    f"Splash screen {splash_screen_filepath} has been "
                    f"DISABLED in {qgiscustomization3ini_filepath}"
                )
                return True
            elif actual_state == nrm_splash_screen_folder and switch:
                logger.debug(
                    f"Splash screen {splash_screen_filepath} is already "
                    f"ENABLED in {qgiscustomization3ini_filepath}"
                )
                return True
            elif actual_state != nrm_splash_screen_folder and switch:
                ini_file.set(
                    section=section,
                    option=option,
                    value=nrm_splash_screen_folder,
                )
                with qgiscustomization3ini_filepath.open(
                    "w", encoding="UTF8"
                ) as configfile:
                    ini_file.write(configfile, space_around_delimiters=False)
                logger.debug(
                    f"Splash screen {splash_screen_filepath} has been "
                    f"ENABLED in {qgiscustomization3ini_filepath}"
                )
                return True
            else:
                pass

        elif ini_file.has_section(section=section) and switch:
            # section exist but not the option, so let's add it as required
            ini_file.set(
                section=section,
                option=option,
                value=nrm_splash_screen_folder,
            )
            with qgiscustomization3ini_filepath.open(
                "w", encoding="UTF8"
            ) as configfile:
                ini_file.write(configfile, space_around_delimiters=False)
            logger.debug(
                f"'{option}' option was missing in {section}, it's just been added and "
                f"ENABLED in {qgiscustomization3ini_filepath}"
            )
            return True
        elif not ini_file.has_section(section=section) and switch:
            # even the section is missing. Let's add everything
            ini_file.add_section(section)
            ini_file.set(
                section=section,
                option=option,
                value=nrm_splash_screen_folder,
            )
            with qgiscustomization3ini_filepath.open(
                "w", encoding="UTF8"
            ) as configfile:
                ini_file.write(configfile, space_around_delimiters=False)
            logger.debug(
                f"'{option}' option was missing in {section}, it's just been added and "
                f"ENABLED in {qgiscustomization3ini_filepath}"
            )
            return True
        else:
            logger.debug(
                f"Section '{section}' is not present so "
                f"{option} is DISABLED in {qgiscustomization3ini_filepath}"
            )
            return False

    @staticmethod
    def _copy_section(
        config_dest: CustomConfigParser,
        config_src: CustomConfigParser,
        section_src: str,
    ) -> None:
        """Copy section from an INI config to another INI config

        Args:
            config_dest (CustomConfigParser): destination INI content
            config_src (CustomConfigParser): source INI content
            section_src (str): section to copy
        """
        if section_src not in config_dest.sections() and section_src != "DEFAULT":
            config_dest.add_section(section_src)
        for param in config_src[section_src]:
            config_dest[section_src][param] = config_src[section_src][param]

    @staticmethod
    def _backup_section(
        config_dest: CustomConfigParser,
        config_src: CustomConfigParser,
        section_src: str,
        dst: Path,
    ) -> None:
        """Backup a INI section for with updated values

        Args:
            config_dest (CustomConfigParser): destination INI content
            config_src (CustomConfigParser): source INI content
            section_src (str): section to backup
            dst (Path): destination file
        """
        if section_src not in config_dest:
            return
        # Get updated values
        updated_values = {
            param: config_dest[section_src][param]
            for param in config_src[section_src]
            if param in config_dest[section_src]
            and config_src[section_src][param] != config_dest[section_src][param]
        }
        if len(updated_values):
            backup_section = f"QDT_backup_{section_src}"
            logger.info(
                f"Section {section_src} already available in {dst}. Copying updated content to {backup_section}"
            )
            if backup_section not in config_dest.sections():
                config_dest.add_section(backup_section)
            for param, backup_val in updated_values.items():
                config_dest[backup_section][param] = backup_val

    def merge_to(self, dst: QgisIniHelper) -> None:
        """Merge INI file to another INI file.
        If the destination file exists a merge is done:
        - all available sections are kept
        - if a section is available in both INI files, keep updated parameters in a backup section
        If environnement variable interpolation is enabled, value are written with current environnement values

        Args:
            dst (QgisIniHelper): destination ini file
        """
        if not self.ini_filepath or not self.ini_filepath.exists():
            logger.warning(
                f"File {self.ini_filepath} doesn't exists. Can't merge to {dst}"
            )
            return

        # Read source INI content
        config_src = self.cfg_parser()
        config_src.read(self.ini_filepath)

        if dst.ini_filepath.exists():
            # Read destination INI content
            config_dest = dst.cfg_parser()
            config_dest.read(dst.ini_filepath)
            # Add sections in source INI
            for section in config_src:
                self._backup_section(config_dest, config_src, section, dst.ini_filepath)
                self._copy_section(config_dest, config_src, section)
            # Write to destination, environnement variable will be interpolated if interpolation enabled
            with dst.ini_filepath.open(mode="w", encoding="UTF-8") as config_file:
                config_dest.write(config_file)
        else:
            with dst.ini_filepath.open(mode="w", encoding="UTF-8") as config_file:
                config_src.write(config_file)


# #############################################################################
# ##### Stand alone program ########
# ##################################

if __name__ == "__main__":
    """Standalone execution."""
    pass



#! python3  # noqa E265

"""
    Usage from the repo root folder:

    .. code-block:: bash
        # for whole tests
        python -m unittest tests.test_rules_context
        # for specific test
        python -m unittest tests.test_rules_context.testQdtRulesContext.test_rules_export_to_json
"""


# standard library
import json
import tempfile
import unittest
from pathlib import Path

# project
from qgis_deployment_toolbelt.profiles.rules_context import QdtRulesContext

# ############################################################################
# ########## Classes #############
# ################################


class TestQdtRulesContext(unittest.TestCase):
    """Test QDT rules context."""

    def test_rules_export_to_json(self):
        """Test export to JSON."""

        rules_context = QdtRulesContext()

        with tempfile.TemporaryDirectory(
            prefix="qdt_test_rules_context"
        ) as tmp_dir_name:
            context_json_path = Path(tmp_dir_name).joinpath("qdt_rules_context.json")

            # write into the file passing extra parameters to json.dumps
            with context_json_path.open("w", encoding="UTF8") as wf:
                wf.write(rules_context.to_json(indent=4, sort_keys=True))

            # test reading
            with context_json_path.open(mode="r", encoding="utf8") as in_json:
                context_data = json.load(in_json)

        self.assertIsInstance(context_data, dict)
        self.assertIn("date", context_data)
        self.assertIn("environment", context_data)
        self.assertIn("user", context_data)


# ############################################################################
# ####### Stand-alone run ########
# ################################
if __name__ == "__main__":
    unittest.main()



#! python3  # noqa: E265

"""
    Read and validate scenario files.

    Author: Julien Moura (https://github.com/guts, Oslandia)
"""


# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import logging
from io import BufferedIOBase
from pathlib import Path

# 3rd party
import yaml

# package
from qgis_deployment_toolbelt.utils.check_path import check_path

# #############################################################################
# ########## Globals ###############
# ##################################

# logs
logger = logging.getLogger(__name__)


# #############################################################################
# ########## Classes ###############
# ##################################


class ScenarioReader:
    """Read and validate scenario files."""

    scenario: dict | None = None

    def __init__(self, in_yaml: str | Path | BufferedIOBase):
        """Instanciajeting YAML scenario reader.

        Args:
            in_yaml (str | Path | BufferedIOBase): input YAML file. Can be a path
                (as str or pathlib.Path) or an opened file.

        Raises:
            TypeError: unsupported
        """
        # check and get YAML path
        if isinstance(in_yaml, (str, Path)):
            self.input_yaml = self.check_yaml_file(in_yaml)
            # extract data from input file
            with self.input_yaml.open(mode="r", encoding="UTF-8") as bytes_data:
                self.scenario = yaml.safe_load(bytes_data)
        elif isinstance(in_yaml, BufferedIOBase):
            self.input_yaml = self.check_yaml_buffer(in_yaml)
            # extract data from input file
            self.scenario = yaml.safe_load(self.input_yaml)
        else:
            raise TypeError(
                "Unsupported data type. Expects a str, Path or buffered IO. "
                f"Got: {type(in_yaml)}"
            )

    # -- CHECKS --
    def check_yaml_file(self, yaml_path: str | Path) -> Path:
        """Perform some checks on passed yaml file and load it as Path object.

        Args:
            yaml_path (str | Path): path to the yaml file to check

        Raises:
            yaml.YAMLError: if YAML is not readable or invalid
            Exception: unhandled error

        Returns:
            Path: sanitized yaml path
        """
        # if path as string load it in Path object
        check_path(
            input_path=yaml_path,
            must_be_a_file=True,
            must_exists=True,
            must_be_readable=True,
        )
        yaml_path = Path(yaml_path)

        # check integrity and structure
        with yaml_path.open(mode="r", encoding="UTF-8") as in_yaml_file:
            try:
                yaml.safe_load_all(in_yaml_file)
            except yaml.YAMLError as exc:
                logger.error(msg=f"YAML file is invalid: {yaml_path.resolve()}")
                raise exc
            except Exception as exc:
                logger.error(
                    f"Something went wrong when reading the scenario file: {yaml_path}. "
                    f"Maybe the structure of YAML file is incorrect? Trace: {exc}"
                )
                raise exc

        # return sanitized path
        return yaml_path

    def check_yaml_buffer(self, yaml_buffer: BufferedIOBase) -> BufferedIOBase:
        """Perform some checks on passed yaml file.

        Args:
            yaml_buffer (BufferedIOBase): bytes reader of the yaml file to check

        Raises:
            yaml.YAMLError: if YAML is not readable or invalid
            Exception: unhandled error

        Returns:
            BufferedIOBase: checked bytes object
        """

        # check integrity
        try:
            yaml.safe_load_all(yaml_buffer)
        except yaml.YAMLError as exc:
            logger.error(f"Invalid YAML {yaml_buffer}. Trace: {exc}")
            raise exc
        except Exception as exc:
            logger.error(
                f"Something went wrong when reading the scenario file. "
                f"Maybe the structure of YAML file is incorrect? Trace: {exc}"
            )
            raise exc

        # return sanitized path
        return yaml_buffer

    def validate_scenario(self) -> tuple[bool, list[str] | None]:
        """Validate scenario file.

        Returns:
            True if scenario is valid, False otherwise and a
            report of validation errors (which is None if the scenario is valid).
        """
        # variables
        required_root_keys: tuple = ("metadata", "settings", "steps")

        # outputs
        valid: bool = True
        report: list[str] = []

        # check if scenario is a dict
        if not isinstance(self.scenario, dict):
            report.append(f"Scenario is not a dict but {type(self.scenario)}")
            valid = False

        # check scenario basic structure
        if any([i not in self.scenario for i in required_root_keys]):
            report.append(
                "Some of required root keys are missing: {}".format(
                    ", ".join(required_root_keys)
                )
            )
            valid = False

        # check if metadata is a dict
        if not isinstance(self.metadata, dict):
            report.append(f"Metadata is not a dict: {self.metadata}")
            valid = False

        return valid, report

    @property
    def metadata(self) -> dict | None:
        """Get metadata from scenario.

        Returns:
            metadata if it does exist
        """
        if isinstance(self.scenario, dict):
            return self.scenario.get("metadata")

    @property
    def settings(self) -> dict | None:
        """Get scenario settings from scenario.

        Returns:
            settings section if it does exist
        """
        if isinstance(self.scenario, dict):
            return self.scenario.get("settings")

    @property
    def steps(self) -> list[dict] | None:
        """Get steps from scenario.

        Returns:
            list of scenario's steps
        """
        if isinstance(self.scenario, dict):
            return self.scenario.get("steps")
